""" Time Conversion Module
Convert between solar time and clock time, given a known longitude. This is the
definition of solar time and standard time, given in equation (1.5.2) in [1].

    [1] Duffie, John A., and William A. Beckman. Solar engineering of thermal
        processes. New York: Wiley, 1991.
"""

from pvsystemprofiler.utilities.equation_of_time import eot_da_rosa, eot_duffie


def solar_to_clock(solar_time, lon, doy, gmt_offset, eot='duffie'):
    """
    :param solar_time: solar time in minutes since midnight (float or array)
    :param lon: longitude (float)
    :param doy: day of year (float or array)
    :param gmt_offset: local timezone offset in hours from UTC/GMT (float or int)
    :param eot: string specifying which equation of time formulation to use
    :return:
    """
    if eot.lower() in ('duffie', 'd'):
        eot = eot_duffie(doy)
    elif eot.lower() in ('da_rosa', 'dr'):
        eot = eot_da_rosa(doy)
    else:
        print('Please select either Duffie or Da Rosa for the equation of time')
        return
    st = solar_time
    ct = st - eot - 4 * (lon - 15 * gmt_offset)
    return ct


def clock_to_solar(clock_time, lon, doy, gmt_offset, eot='duffie'):
    if eot.lower() in ('duffie', 'd'):
        eot = eot_duffie(doy)
    elif eot.lower() in ('da_rosa', 'dr'):
        eot = eot_da_rosa(doy)
    else:
        print('Please select either Duffie or Da Rosa for the equation of time')
        return
    ct = clock_time
    st = ct + eot + 4 * (lon - 15 * gmt_offset)
    return st



import os
import sys
import json
import boto3
import subprocess
import paramiko
from smart_open import smart_open
import numpy as np
import pandas as pd
from solardatatools.utilities import progress
from pvsystemprofiler.longitude_study import LongitudeStudy
from pvsystemprofiler.latitude_study import LatitudeStudy
from pvsystemprofiler.tilt_azimuth_study import TiltAzimuthStudy


def get_address(tag_name, region, client):
    """
    Collects the addresses of the aws instances being used for the estimation
    :param tag_name: aws 'Name' tag of the instances
    :param region: aws region
    :param client: aws client
    :return: list with aws instance addresses
    """
    ec2 = boto3.Session(profile_name='default', region_name=region).client(client)
    target_instances = ec2.describe_instances(Filters=[{'Name': 'tag:Name', 'Values': [tag_name]}])

    ec2_instances = []
    for each_instance in target_instances['Reservations']:
        for found_instance in each_instance['Instances']:
            if found_instance['PublicDnsName'] != '':
                ec2_instances.append(found_instance['PublicDnsName'])
    return ec2_instances


def remote_execute(user, instance_id, key, shell_commands, verbose=True):
    """
    Executes a list of bash commands remotely on Amazon Web Services (AWS) instances from another computer.
    :param user: AWS instance user name, i.e. `ubuntu`.
    :param instance_id: AWS public ip address of instance.
    :param key: AWS key. Usually a *.pem file located in the .aws folder.
    :param shell_commands: list of shell commands to be executed remotely.
    :param verbose: provides the output to each remote command execution.
    """
    k = paramiko.RSAKey.from_private_key_file(key)
    c = paramiko.SSHClient()
    c.set_missing_host_key_policy(paramiko.AutoAddPolicy())
    c.connect(hostname=instance_id, username=user, pkey=k, allow_agent=False, look_for_keys=False)
    command_dict = {}
    for command in shell_commands:
        if verbose:
            print("running command: {}".format(command))
        stdin, stdout, stderr = c.exec_command(command)
        command_dict[command] = [stdout.read(), stderr.read()]
    c.close()
    return command_dict


def copy_to_s3(input_file_name, input_file_location):
    """
    Copies a local file to a Amazon Web Services (AWS) s3 bucket.
    :param input_file_name: name of the file to be copied to the AWS s3 bucket.
    :param input_file_location: full path to the destination AWS s3 bucket.
    """
    bucket, prefix = get_s3_bucket_and_prefix(input_file_location)
    destination_file_name = prefix + '/generated_site_list.csv'

    content = open(input_file_name, 'rb')
    s3 = boto3.client('s3')
    s3.put_object(Bucket=bucket, Key=destination_file_name, Body=content)


def get_s3_bucket_and_prefix(s3_location):
    """
    Splits absolute s3 location into parameters used to copy and execute commands remotely.
    :param s3_location: full path to s3 bucket location.
    :return: s3 bucket and prefix.
    """
    if s3_location[-1] != '/':
        s3_location += '/'
    i = 5
    j = s3_location.find('/', i)
    bucket = s3_location[i:j]
    prefix = s3_location[j + 1:-1]
    return bucket, prefix


def load_generic_data(location, file_label, file_id, extension='.csv', parse_dates=[0], nrows=None):
    """
    Loads csv file containing input signals for a given site.
    :param location: String. absolute path to csv file containing input signals.
    :param file_label: String. Repeating portion of data files label. If 'None', no file label is used.
    :param file_id: String. Individual identifier of a site csv file
    :param extension: String, optional. Extension of file containing input signal.
    :param parse_dates: Optional. 'read_csv' kwarg.
    :param nrows: number of rows from input signal file to be read.
    :return: Dataframe containing input signals for `file_id`.
    """
    if file_label is None:
        to_read = location + file_id + extension
    else:
        to_read = location + file_id + file_label + extension

    if nrows is None:
        df = pd.read_csv(to_read, index_col=0, parse_dates=parse_dates)
    else:
        df = pd.read_csv(to_read, index_col=0, parse_dates=parse_dates, nrows=nrows)
    return df


def get_checked_sites(df):
    """
    Returns list of sites that have already been analyzed
    :param df: pandas dataframe containing results from reports or parameter study.
    :return:  List with sites that have already been analyzed.
    """
    if df is not None:
        checked_sites = df['site'].unique().tolist()
        checked_sites.sort()
    else:
        checked_sites = []
    return checked_sites


def resume_run(results_file):
    """
    Loads the output dataFrame from an incomplete run provided the results csv file name
    :param results_file: full path to csv file containing partial results.
    :return: df dataFrame containing partial results, None if there are no partial results.
    """
    if os.path.isfile(results_file):
        df = pd.read_csv(results_file, index_col=0)
        df['site'] = df['site'].apply(str)
        df['system'] = df['system'].apply(str)
    else:
        df = None
    return df


def enumerate_files(s3_location, extension='.csv', file_size_list=False):
    """
    Returns a list with the file names with a given extension located in a AWS s3 bucket.
    :param s3_location: String. Full path to s3 bucket from which a list of files is to be generated.
    :param extension: String. Extension of the files to be included in `output_list`.
    :param file_size_list: Boolean, generate a list with the size of each file with `extension`.
    :return: `output_list' with file names and (optional) list with site of files in `output list'.
    """
    s3_bucket, prefix = get_s3_bucket_and_prefix(s3_location)
    s3 = boto3.client('s3')
    output_list = []
    size_list = []
    for obj in s3.list_objects_v2(Bucket=s3_bucket, Prefix=prefix)['Contents']:
        if (obj['Key']).find(extension) != -1:
            file_name = obj['Key']
            file_size = obj['Size']
            i = file_name.rfind('/')
            file_name = file_name[i + 1:]
            output_list.append(file_name)
            size_list.append(file_size)
    if file_size_list:
        return output_list, size_list
    else:
        return output_list


def create_system_dict(df):
    """
    Reads a pandas dataFrame and creates a dictionary where the keys are the site ids and the values are a the list of
    systems for each site.
    :param df: pandas dataFrame with site and system information a `site` and a `system` column.
    :return: dictionary with systems associated to each site.
    """
    site_list = df['site'].unique().tolist()
    ss_dict = {}
    for site in site_list:
        systems_in_site = df[df['site'] == site]['system'].values.tolist()
        ss_dict[site] = systems_in_site
    return site_list, ss_dict


def create_json_dict(json_list, location):
    """
    returns a dictionary containing the system ids given a location containing json files with site information.
    :param json_list: list of json files containing site information.
    :param location: absolute path to folder containing json files with site information.
    :return: dictionary containing system ids as keys and site ids as values.
    """
    system_dict = {}
    for file in json_list:
        for line in smart_open(location + file, 'rb'):
            file_json = json.loads(line)
            if len(file_json['Inverters']) != 0:
                for inv_id in file_json['Inverters']:
                    system = file_json['Inverters'][inv_id]['inverter_id']
                    system_dict[system] = file
    return system_dict


def log_file_versions(utility, active_conda_env=None, output_folder_location='./',
                      conda_location='/home/ubuntu/miniconda3/', repository_location='/home/ubuntu/github/'):
    """
    Writes a text file with the utility version as well as other available package information. This utility may have
    been installed in conda using pip or may be in a local GitHub repository.
    :param utility: name of the utility to be logged. Examples are `solar-data-tools' and 'pv-system-profiler'
    :param active_conda_env: optional. Name of the conda environment being used. Examples are pvi-user and `pvi-dev`
    :param output_folder_location: location of the output log files.
    :param conda_location: optional. Full path to the local conda folder location.
    :param repository_location: optional. Location of the GitHub repository.
    :return:
    """
    if active_conda_env is None:
        conda = conda_location + 'bin/conda' + ' '
        conda_env = subprocess.check_output(conda + "env list", shell=True, encoding='utf-8')
        for line in conda_env.splitlines():
            if '*' in line:
                i = line.find('/')
                active_conda_environment = line[i:].split('/')[-1]
    else:
        active_conda_environment = active_conda_env

    output_string = 'active conda environment:' + ' ' + active_conda_environment + '\n'
    pip = conda_location + 'envs/' + active_conda_environment + '/bin/pip'
    try:
        pip_list = subprocess.check_output(pip + ' ' + 'show' + ' ' + utility, shell=True, encoding='utf-8')

        for line in pip_list.splitlines():
            if 'Location' in line:
                i = line.find(':')
                location = line[i + 1:]
            if 'Version' in line:
                i = line.find(':')
                location = line[i + 1:]
                output_string += 'utility version:' + ' ' + location + '\n'
    except:
        pip_list = ''
        output_string += 'utility version:' + ' ' + 'n/a' + '\n'
    try:
        if pip_list.find('site-packages') != -1:
            repository = ''
        else:
            if repository_location is not None:
                location = repository_location + utility
            repository = subprocess.check_output('/usr/bin/git -C' + ' ' + location + ' ' + 'log -n 1', shell=True,
                                                 encoding='utf-8')
        output_string += 'repository location:' + ' ' + location + '\n'

        for line in repository.splitlines():
            if line.find('commit') != - 1:
                i = line.find(' ')
                commit_id = line[i:]
                output_string += 'commit id:' + ' ' + commit_id + '\n'
            if line.find('Author') != - 1:
                i = line.find(': ')
                author = line[i + 1:]
                output_string += 'author:' + ' ' + author + '\n'
            if line.find('Date') != - 1:
                i = line.find(' ')
                date = line[i:]
                output_string += 'date:' + ' ' + date + '\n'
    except:
        output_string += 'commit id:' + ' ' + 'n/a' + '\n'
        output_string += 'author:' + ' ' + 'n/a' + '\n'
        output_string += 'date:' + ' ' + 'n/a' + '\n'

    output_file = open(output_folder_location + utility + '_' + 'versions.log', 'w')
    output_file.write(output_string)
    output_file.close()
    return


def create_system_list(file_label, signal_label, s3_location):
    """
    returns a list of systems present in a `s3_bucket`.
    :param file_label: String. Repeating part of label of files containing input data. For the site list
    ['1_signal.csv', '2_signal.csv'] with `file_label`='_signal'.
    :param signal_label: String. Label of the input signal, i.e. `ac_power_inv_` and `dc_current_inv`.
    :param s3_location: full path to AWS s3 bucket containing csv files with site input signals.
    :return: List containing ids for systems in s3_location.
    """
    bucket, prefix = get_s3_bucket_and_prefix(s3_location)
    file_list = enumerate_files(bucket, prefix)
    ll = len(signal_label)
    system_list = pd.DataFrame(columns=['site', 'system'])

    for file_ix, file_id in enumerate(file_list):
        progress(file_ix, len(file_list), 'Generating system list', bar_length=20)
        file_name = file_id.split('/')[1]
        i = file_name.find(file_label)
        file_id = file_name[:i]
        df = load_generic_data(s3_location, file_label, file_id, nrows=2)
        cols = df.columns
        for col_label in cols:
            if col_label.find(signal_label) != -1:
                system_id = col_label[ll:]
                system_list.loc[len(system_list)] = file_id, system_id
    progress(len(file_list), len(file_list), 'Generating system list', bar_length=20)
    return system_list


def extract_sys_parameters(file_name, system, location):
    """
    Retrieves a list with zip code, inverter id, longitude, latitude, tilt and azimuth for a given system from
    a json file containing site information.
    site.
    :param file_name: String. Name of the json file containing site information.
    :param system: String. system id for which information is to be retrieved from json file
    :param location: String. Full path to folder containing site json file.
    :return: list with system zip code, inverter id, longitude, latitude, tilt and azimuth
    """
    for line in smart_open(location + file_name, 'rb'):
        file_json = json.loads(line)
        parameters = []
        if len(file_json['Inverters']) == 0:
            parameters = [np.nan] * 5
        else:
            try:
                zc = file_json['Site']['location'][-5:]
                zc = '00000' if not zc.isnumeric() else zc
            except KeyError:
                zc = '00000'
            parameters.append(zc)
            for inv_id in file_json['Inverters']:
                mount_id = 'Mount ' + inv_id.split(' ')[1]
                sys_id = file_json['Inverters'][inv_id]['inverter_id']
                if sys_id == system:
                    for coord_ix, coord_id in enumerate(['longitude', 'latitude', 'tilt', 'azimuth']):
                        try:
                            if coord_id in ['longitude', 'latitude']:
                                val = float(file_json['Site'][coord_id])
                            if coord_id in ['tilt', 'azimuth']:
                                val = file_json['Mount'][mount_id][coord_id]
                        except KeyError:
                            val = np.nan
                        val = np.nan if val == '' else val
                        parameters.append(val)
        #file_id = file_json['System']['system_id']
        #parameters.append(file_id)
        return parameters


def siteid_to_filename(sites, file_label, ext='csv'):
    """
    Given a list of site ids returns a list of strings corresponding to the file name obtained by concatenating it to
    the 'file_label` string.  For `sites` in ['1', '2'] with `file_label`='_signal', this method will return the list
    ['1_signal.csv', '2_signal.csv'].
    :param sites: List of strings containing the file id, i.e. non-repeating part of the input signal csv file names.
    :param file_label: String. Repeating part of the input signal csv file name.
    :param ext: String, optional. Input signal file extension
    :return: full file name corresponding to site ids in `sites` list
    """
    checked_sites = []
    for site_id in sites:
        file_name = site_id + file_label + ext
        checked_sites.append(file_name)
    return checked_sites


def filename_to_siteid(sites):
    """
    Given a list of file names containing input signals returns a list of strings with the site ids only. For the list
    of file ids `sites` in ['1_signal.csv', '2_signal.csv'] with `file_label`='_signal', this method will return a list
    with the site ids ['1', '2'].
    .
    :param sites: List of strings containing the full name of csv files containing input signals.
    :param file_label: String. Repeating part of the input signal csv file name.
    :return: list of site ids.
    """
    site_list = []
    for site_id in sites:
        site_id = site_id.split('.')[0]
        site_list.append(site_id)
    return site_list


def run_failsafe_pipeline(dh, sys_tag, fts, tzc):
    """
    Runs the solarDataTools dataHandler pipeline in failsafe mode.
    :param dh: input data handler
    :param sys_tag: Dataframe column label identifying an input signal, i.e. ac_power_01 ar dc_current_02.
    :param fts: Boolean. Fix time shift parameter in `run_pipeline`.
    :param tzc: Boolean. Time zone correction parameter in `run_pipeline`
    :return: Boolean. True if passes pipeline, otherwise False.
    """
    df = dh.data_frame_raw
    try:
        try:
            dh.run_pipeline(power_col=sys_tag, fix_shifts=fts, correct_tz=tzc, verbose=False)
        except ValueError:
            max_val = np.nanquantile(df[sys_tag], 0.95)
            dh.run_pipeline(power_col=sys_tag, fix_shifts=False, correct_tz=tzc, verbose=False, max_val=max_val * 3)
    except:
        return dh, False
    return dh, True


def get_commandline_inputs(input_kwargs):
    """
        :param estimation: Estimation to be performed. Options are 'report', 'longitude', 'latitude', 'tilt_azimuth'.
        :param input_site_file:  csv file containing list of sites to be evaluated. 'None' if no input file is provided.
        :param n_files: number of files to read. If 'all' all files in folder are read.
        :param s3_location: Absolute path to s3 location of files.
        :param file_label:  Repeating portion of data files label. If 'None', no file label is used.
        :param power_column_label: Repeating portion of the power column label.
        :param output_file: Absolute path to csv file containing report results.
        :param fix_time_shits: String, 'True' or 'False'. Specifies if time shifts are to be
        fixed when running the pipeline.
        :param time_zone_correction: String, 'True' or 'False'. Specifies if the time zone correction is performed when
        running the pipeline.
        :param check_json: String, 'True' or 'False'. Check json file for location information.
        :param convert_to_ts: String, 'True' or 'False'. Specifies if conversion to time series is performed when
        running the pipeline.
        :param system_summary_file: Full path to csv file containing longitude and manual time shift flag for each system,
        None if no file
        provided.
        :param gmt_offset: String. Single value of gmt offset to be used for all estimations. If None a list with individual
        gmt offsets needs to be provided.
        :param data_source: String. Input signal data source. Options are 's3' and 'cassandra'.
        """
    inputs_dict = {'estimation': input_kwargs[1],
                   'input_site_file': input_kwargs[2] if input_kwargs[2] != 'None' else None,
                   'n_files': input_kwargs[3],
                   's3_location': input_kwargs[4] if input_kwargs[4] != 'None' else None,
                   'file_label': input_kwargs[5] if input_kwargs[5] != 'None' else None,
                   'power_column_label': input_kwargs[6],
                   'output_file': input_kwargs[7],
                   'fix_time_shifts': True if input_kwargs[8] == 'True' else False,
                   'time_zone_correction': True if input_kwargs[9] == 'True' else False,
                   'check_json': True if input_kwargs[10] == 'True' else False,
                   'convert_to_ts': True if input_kwargs[11] == 'True' else False,
                   'system_summary_file': input_kwargs[12] if input_kwargs[12] != 'None' else None,
                   'gmt_offset': input_kwargs[13] if input_kwargs[13] != 'None' else None,
                   'data_source': input_kwargs[14]}
    return inputs_dict


def load_system_metadata(df_in, file_label):
    df = pd.read_csv(df_in, index_col=0)
    df = df[~df['time_shift_manual'].isnull()]
    df['time_shift_manual'] = df['time_shift_manual'].apply(int)
    df = df[df['time_shift_manual'].isin([0, 1])]
    df['site'] = df['site'].apply(str)
    df['system'] = df['system'].apply(str)
    if file_label is not None:
        df['site_file'] = df['site'].apply(lambda x: str(x) + file_label)
    return df


def generate_list(inputs_dict, full_df, df_system_metadata):
    if inputs_dict['s3_location'] is not None:
        full_site_list = enumerate_files(inputs_dict['s3_location'])
        full_site_list = filename_to_siteid(full_site_list)
    else:
        full_site_list = []

    previously_checked_site_list = get_checked_sites(full_df)
    file_list = list(set(full_site_list) - set(previously_checked_site_list))

    if inputs_dict['check_json']:
        json_files = enumerate_files(inputs_dict['s3_location'], extension='.json')
        print('Generating system list from json files')
        json_file_dict = create_json_dict(json_files, inputs_dict['s3_location'])
        print('List generation completed')
    else:
        json_file_dict = None

    if inputs_dict['input_site_file'] is not None:
        input_site_list_df = pd.read_csv(inputs_dict['input_site_file'], index_col=0)
        site_list = input_site_list_df['site'].apply(str)
        site_list = site_list.tolist()
        if len(file_list) != 0:
            file_list = list(set(site_list) & set(file_list))
        else:
            file_list = list(set(site_list))
        if inputs_dict['time_shift_manual']:
            manually_checked_sites = df_system_metadata['site_file'].apply(str).tolist()
            file_list = list(set(file_list) & set(manually_checked_sites))
    file_list.sort()

    return file_list, json_file_dict


def run_failsafe_lon_estimation(dh_in, real_longitude, gmt_offset):
    try:
        runs_lon_estimation = True
        lon_study = LongitudeStudy(data_handler=dh_in, gmt_offset=gmt_offset, true_value=real_longitude)
        lon_study.run(verbose=False)
        p_df = lon_study.results.sort_index().copy()
    except:
        runs_lon_estimation = False
        p_df = pd.DataFrame(columns=['longitude', 'estimator', 'eot_calculation', 'solar_noon_method',
                                     'day_selection_method', 'data_matrix', 'residual', 'site', 'system'])
        p_df.loc[0, :] = np.nan
    return p_df, runs_lon_estimation


def run_failsafe_lat_estimation(dh_in, real_latitude):
    try:
        runs_lat_estimation = True
        lat_study = LatitudeStudy(data_handler=dh_in, lat_true_value=real_latitude)
        lat_study.run()
        p_df = lat_study.results.sort_index().copy()
    except:
        runs_lat_estimation = False
        p_df = pd.DataFrame(columns=['declination_method', 'daylight_calculation', 'data_matrix', 'threshold',
                                     'day_selection_method', 'latitude', 'residual'])
        p_df.loc[0, :] = np.nan
    return p_df, runs_lat_estimation


def run_failsafe_ta_estimation(dh, nrandom, threshold, lon, lat, tilt, azim, real_lat, real_tilt, real_azim,
                               gmt_offset):
    try:
        runs_ta_estimation = True
        ta_study = TiltAzimuthStudy(data_handler=dh, nrandom_init_values=nrandom, daytime_threshold=threshold,
                                    lon_input=lon, lat_input=lat, tilt_input=tilt, azimuth_input=azim,
                                    lat_true_value=real_lat, tilt_true_value=real_tilt, azimuth_true_value=real_azim,
                                    gmt_offset=gmt_offset)
        ta_study.run()
        p_df = ta_study.results.sort_index().copy()
    except:
        runs_ta_estimation = False
        cols = ['day range', 'declination method', 'latitude initial value', 'tilt initial value',
                'azimuth initial value']
        if lat is not None:
            cols.append('latitude')
        if tilt is not None:
            cols.append('tilt')
        if azim is not None:
            cols.append('azimuth')
        if lat is not None:
            cols.append('latitude_residual')
        if tilt is not None:
            cols.append('tilt_residual')
        if azim is not None:
            cols.append('azimuth_residual')
        p_df = pd.DataFrame(columns=cols)
        p_df.loc[0, :] = np.nan
    return p_df, runs_ta_estimation



import numpy as np
import pandas as pd


def random_initial_values(nrandom):
    lat_initial_value = np.random.uniform(low=-90, high=90, size=nrandom)
    tilt_initial_value = np.random.uniform(low=0, high=90, size=nrandom)
    azim_initial_value = np.random.uniform(low=-180, high=180, size=nrandom)
    return lat_initial_value, tilt_initial_value, azim_initial_value



""" Latitude Study Module
This module contains a class for conducting a study to estimate latitude from solar data. This code accepts solar data
in the form of a `solar-data-tools` `DataHandler` object, which is used to standardize and pre-process the data. The
provided class will then estimate the latitude of the site that produced the data, using the `run` method.

The following configurations can be run:

 - Input data matrix: 'raw', 'filled'
 - Daylight estimation method: 'raw daylight', 'sunrise-sunset', 'optimized_estimates', 'optimized_measurements'
 - Declination equation: 'cooper', 'spencer'
 - Day selection method: 'all', 'clear', 'cloudy'

"""
import numpy as np
import pandas as pd
from pvsystemprofiler.utilities.declination_equation import delta_spencer
from pvsystemprofiler.utilities.declination_equation import delta_cooper
from pvsystemprofiler.algorithms.latitude.hours_daylight import calculate_hours_daylight
from pvsystemprofiler.algorithms.latitude.hours_daylight import calculate_hours_daylight_raw
from pvsystemprofiler.algorithms.optimized_sunrise_sunset import get_optimized_sunrise_sunset
from pvsystemprofiler.algorithms.latitude.estimation import estimate_latitude


class LatitudeStudy():
    def __init__(self, data_handler, lat_true_value=None):
        """
        :param data_handler: `DataHandler` class instance loaded with a solar power data set.
        :param lat_true_value: Optional. The ground truth value for the system's latitude. (Degrees).
        """

        self.data_handler = data_handler
        self.latitude_true_value = lat_true_value
        if not data_handler._ran_pipeline:
            print('Running DataHandler preprocessing pipeline with defaults')
            self.data_handler.run_pipeline()
        self.data_matrix = self.data_handler.filled_data_matrix
        self.raw_data_matrix = self.data_handler.raw_data_matrix
        self.day_of_year = self.data_handler.day_index.dayofyear
        self.num_days = self.data_handler.num_days
        self.daily_meas = self.data_handler.filled_data_matrix.shape[0]
        self.data_sampling = self.data_handler.data_sampling
        self.boolean_daytime = None
        self.hours_daylight = None
        self.delta = None
        self.delta_cooper = None
        self.delta_spencer = None
        self.residual = None
        self.daytime_threshold = None
        self.opt_threshold = None
        self.days = None
        self.estimates_sunrise_raw = None
        self.estimates_sunset_raw = None
        self.measurements_sunrise_raw = None
        self.measurements_sunset_raw = None
        self.estimates_sunrise_filled = None
        self.estimates_sunset_filled = None
        self.measurements_sunrise_filled = None
        self.measurements_sunset_filled = None
        self.opt_threshold_filled = None
        self.opt_threshold_raw = None
        # Results
        self.results = None

    def run(self, data_matrix=('raw', 'filled'),
            daylight_method=('raw daylight', 'sunrise-sunset', 'optimized_estimates', 'optimized_measurements'),
            delta_method=('cooper', 'spencer'), day_selection_method=('all', 'clear', 'cloudy'),
            threshold=None):
        """
        Run a study with the given configuration of options. Defaults to
        running all available options. Any kwarg can be constrained by
        providing a subset of acceptable keys. For example the default keys
        for daylight method kwarg are:

            ('raw daylight', 'sunrise-sunset', 'optimized_estimates', 'optimized_measurements')

        Additionally, any of the following would be acceptable for this kwarg:

            ('raw daylight', 'sunrise-sunset', 'optimized_estimates', 'optimized_measurements')
            ('raw daylight', 'sunrise-sunset', 'optimized_estimates')
            ('raw daylight', 'sunrise-sunset')
            ('raw daylight')

        This method sets the `results` attribute to be a pandas data frame
        containing the results of the study.
        :param data_matrix: 'raw', 'filled'.
        :param daylight_method: 'raw daylight', 'sunrise-sunset', 'optimized_estimates', 'optimized_measurements'.
        :param threshold: (optional) daylight threshold values, tuple of length one to twelve.
        :param delta_method: (optional) 'cooper', 'spencer'.
        :param day_selection_method: 'all', 'clear', 'cloudy'.
        :return: None.
        """
        data_matrix = np.atleast_1d(data_matrix)
        daylight_method = np.atleast_1d(daylight_method)
        delta_method = np.atleast_1d(delta_method)
        day_selection_method = np.atleast_1d(day_selection_method)

        if threshold is None:
            self.daytime_threshold = 0.001 * np.ones(len(data_matrix) * len(daylight_method) * len(delta_method) *
                                                     len(day_selection_method))
        else:
            self.daytime_threshold = threshold

        self.delta_cooper = delta_cooper(self.day_of_year, self.daily_meas)
        self.delta_spencer = delta_spencer(self.day_of_year, self.daily_meas)

        if 'raw' in data_matrix:
            rdm = self.raw_data_matrix
        else:
            rdm = None
        if 'filled' in data_matrix:
            fdm = self.data_matrix
        else:
            fdm = None

        opt_dict = get_optimized_sunrise_sunset(fdm, rdm)
        self.estimates_sunrise_raw, self.estimates_sunset_raw, self.measurements_sunrise_raw, \
        self.measurements_sunset_raw, self.opt_threshold_raw, \
        self.estimates_sunrise_filled, self.estimates_sunset_filled, self.measurements_sunrise_filled, \
        self.measurements_sunset_filled, self.opt_threshold_filled = opt_dict.values()

        results = pd.DataFrame(columns=['declination_method', 'daylight_calculation', 'data_matrix', 'threshold',
                                        'day_selection_method', 'latitude'])
        counter = 0
        for delta_id in delta_method:
            for matrix_ix, matrix_id in enumerate(data_matrix):

                for daylight_method_id in daylight_method:
                    if daylight_method_id != 'optimized_estimates':
                        dtt = self.daytime_threshold[counter]
                    else:
                        dtt = None

                    for ds in day_selection_method:
                        if ds == 'all':
                            self.days = self.data_handler.daily_flags.no_errors
                        elif ds == 'clear':
                            self.days = self.data_handler.daily_flags.clear
                        elif ds == 'cloudy':
                            self.days = self.data_handler.daily_flags.cloudy
                        dlm = daylight_method_id
                        tm = data_matrix[matrix_ix]
                        dm = delta_id

                        self.prepare_input_data(matrix_id, daytime_threshold=dtt, daylight_method=dlm,
                                                delta_method=delta_id)
                        lat_est = estimate_latitude(self.hours_daylight, self.delta)

                        if daylight_method_id in ['optimized_estimates', 'optimized_measurements']:
                            dtt = self.opt_threshold

                        results.loc[counter] = [dm, dlm, tm, dtt, ds, lat_est]
                        counter += 1
        if self.latitude_true_value is not None:
            results['residual'] = self.latitude_true_value - results['latitude']
            results['measured_latitude'] = self.latitude_true_value

        self.results = results

    def prepare_input_data(self, matrix_id=None, daytime_threshold=0.001,
                           daylight_method=('sunrise-sunset', 'raw daylight'),
                           delta_method=('cooper', 'spencer')):
        """"
        Latitude is estimated from equation (1.6.11) in:
        Duffie, John A., and William A. Beckman. Solar engineering of thermal processes. New York: Wiley, 1991.
        """

        if matrix_id == 'raw':
            data_in = self.raw_data_matrix
        elif matrix_id == 'filled':
            data_in = self.data_matrix
        if daylight_method in ('sunrise-sunset', 'sunrise sunset'):
            hours_daylight_all = calculate_hours_daylight(data_in, daytime_threshold)
        elif daylight_method in ('raw_daylight', 'raw daylight'):
            hours_daylight_all = calculate_hours_daylight_raw(data_in, self.data_sampling, daytime_threshold)
        elif daylight_method in ('optimized_estimates', 'Optimized_Estimates'):
            if matrix_id == 'filled':
                hours_daylight_all = self.estimates_sunset_filled - self.estimates_sunrise_filled
                self.opt_threshold = self.opt_threshold_raw
            if matrix_id == 'raw':
                hours_daylight_all = self.estimates_sunset_raw - self.estimates_sunrise_raw
                self.opt_threshold = self.opt_threshold_filled
        elif daylight_method in ('optimized_measurements', 'Optimized_Measurements'):
            if matrix_id == 'filled':
                hours_daylight_all = self.measurements_sunset_filled - self.measurements_sunrise_filled
                self.opt_threshold = self.opt_threshold_filled
            if matrix_id == 'raw':
                hours_daylight_all = self.measurements_sunset_raw - self.measurements_sunrise_raw
                self.opt_threshold = self.opt_threshold_raw
        if delta_method in ('Cooper', 'cooper'):
            self.delta = self.delta_cooper
        elif delta_method in ('Spencer', 'spencer'):
            self.delta = self.delta_spencer

        if np.any(np.isnan(hours_daylight_all)):
            hours_mask = np.isnan(hours_daylight_all)
            full_mask = ~hours_mask & self.days
            self.hours_daylight = hours_daylight_all[full_mask]
            self.delta = self.delta[:, full_mask]
        else:
            self.hours_daylight = hours_daylight_all[self.days]
            self.delta = self.delta[:, self.days]
        return



"""
This module is used to set the hour_angle_equation in terms of the unknowns. The hour equation is a function of the
declination (delta), the hour angle (omega) , latitude (phi), tilt (beta) and azimuth (gamma). The declination and the
hour angle are treated as input parameters for all cases. Latitude, tilt and azimuth can be given as input parameters
 or left as unknowns (`None`). In total, seven different combinations arise from having these three parameters
as an inputs or as a unknowns. The seven conditionals below correspond to those combinations. The output function `func`
is used as one of the inputs to run_curve_fit which in turn is used to fit the unknowns. The function other outputs is
the 'bounds' tuple containing the bounds for the variables. Bounds for latitude are -90 to 90. Bounds for tilt are 0 to
90. Bounds for azimuth  are -180 to 180. It is noted that, theoretically, bounds for tilt are 0 to 180 (Duffie, John A.,
 and William A. Beckman. Solar engineering of thermal processes. New York: Wiley, 1991.). However a value of tilt >90
 would mean that that the surface has a downward-facing component, which is not the case of the current application.
"""
from pvsystemprofiler.utilities.angle_of_incidence_function import func_costheta
import numpy as np


def select_function(latitude=None, tilt=None, azimuth=None):
    """
    :param latitude: (optional) latitude input value in Degrees.
    :param tilt: (optional) Tilt input value in Degrees.
    :param azimuth: (optional) Azimuth input value in Degrees.
    :return: Customized function 'func' and 'bounds' tuple.
    """

    if latitude is None and tilt is None and azimuth is None:
        func = lambda x, phi, beta, gamma: func_costheta(x, phi, beta, gamma)
    elif latitude is not None and tilt is None and azimuth is None:
        func = lambda x, beta, gamma: func_costheta(x, np.deg2rad(latitude), beta, gamma)

    elif latitude is None and tilt is not None and azimuth is None:
        func = lambda x, phi, gamma: func_costheta(x, phi, np.deg2rad(tilt), gamma)

    elif latitude is None and tilt is None and azimuth is not None:
        func = lambda x, phi, beta: func_costheta(x, phi, beta, np.deg2rad(azimuth))

    elif latitude is None and tilt is not None and azimuth is not None:
        func = lambda x, phi: func_costheta(x, phi, np.deg2rad(tilt), np.deg2rad(azimuth))

    elif latitude is not None and tilt is None and azimuth is not None:
        func = lambda x, beta: func_costheta(x, np.deg2rad(latitude), beta, np.deg2rad(azimuth))

    elif latitude is not None and tilt is not None and azimuth is None:
        func = lambda x, gamma: func_costheta(x, np.deg2rad(latitude), np.deg2rad(tilt), gamma)

    bounds_dict = {'latitude': [-np.pi / 2, np.pi / 2], 'tilt': [0, np.pi / 2], 'azimuth': [-np.inf, np.inf]}
    bounds = []

    if latitude is None:
        bounds.append(bounds_dict['latitude'])
    if tilt is None:
        bounds.append(bounds_dict['tilt'])
    if azimuth is None:
        bounds.append(bounds_dict['azimuth'])

    bounds = tuple(np.transpose(bounds).tolist())

    return func, bounds



import numpy as np
import pandas as pd
from pvsystemprofiler.utilities.time_convert import solar_to_clock


def sunset_hour_angle(doy, lat):
    b = np.deg2rad((360 / 365) * (doy - 1))
    delta = (0.006918 - 0.399912 * np.cos(b) + 0.070257 * np.sin(b) -
             0.006758 * np.cos(2 * b) + 0.000907 * np.sin(2 * b) -
             0.002697 * np.cos(3 * b) + 0.00148 * np.sin(3 * b))
    sunset_hour_angle = np.arccos(
        -np.tan(np.deg2rad(lat)) * np.tan(delta)
    )
    return np.rad2deg(sunset_hour_angle)


def num_daylight_hours(doy, lat):
    return (2 / 15) * sunset_hour_angle(doy, lat)


def sunrise_sunset_times(lat, lon, doy, gmt_offset, eot='duffie'):
    ss_ha = sunset_hour_angle(doy, lat)
    ss_st = 12 + ss_ha / 15
    sr_st = 12 - ss_ha / 15
    ss_st *= 60
    sr_st *= 60
    ss_ct = solar_to_clock(ss_st, lon, doy, gmt_offset, eot)
    sr_ct = solar_to_clock(sr_st, lon, doy, gmt_offset, eot)
    ss_ct /= 60
    sr_ct /= 60
    output_table = pd.DataFrame(data={
        'day of year': doy,
        'sunrise times': sr_ct,
        'sunset times': ss_ct
    })
    return output_table



import sys
import os
import numpy as np
import pandas as pd
import time
import glob
from pathlib import Path

# TODO: remove pth.append after package is deployed
filepath = Path(__file__).resolve().parents[2]
sys.path.append(str(filepath))
from pvsystemprofiler.scripts.modules.config_partitions import get_config
from pvsystemprofiler.scripts.modules.create_partition import create_partition
from pvsystemprofiler.scripts.modules.script_functions import enumerate_files
from pvsystemprofiler.scripts.modules.script_functions import copy_to_s3
from pvsystemprofiler.scripts.modules.script_functions import remote_execute
from pvsystemprofiler.scripts.modules.script_functions import get_address
from pvsystemprofiler.scripts.modules.script_functions import get_commandline_inputs


def build_input_file(s3_location, input_file_location='s3://pv.insight.misc/report_files/'):
    """
    Builds a csv input file by looking at the contents of the s3 bucket containing csv files with signals.
    :param s3_location: aws s3 bucket location of csv files containing signals
    :param input_file_location: s3 bucket location of report files
    :return: DataFrame with signals in a given folder
    """
    site_list, size_list = enumerate_files(s3_location, file_size_list=True)
    site_df = pd.DataFrame()
    site_df['site'] = site_list
    site_df['site'] = site_df['site'].apply(lambda x: x.split('.')[0])
    site_df['file_size'] = size_list
    site_df.to_csv('./generated_site_list.csv')
    copy_to_s3('./generated_site_list.csv', input_file_location)
    return site_df


def get_remote_output_files(partitions, username, destination_dict):
    """
    Collects partition results once estimation is finished.
    :param partitions: String. List containing aws partition addresses.
    :param username:  String. aws user name.
    :param destination_dict: String. Folder where results are saved.
    """
    os.system('mkdir' + ' ' + destination_dict)
    for part_id in partitions:
        get_local_output_file = "scp -i" + part_id.ssh_key_file + ' ' + username + "@" + part_id.public_ip_address + \
                                ":" + part_id.local_output_file + ' ' + destination_dict
        os.system(get_local_output_file)


def combine_results(partitions, destination_dict):
    """
    Combines partitioned results into single csv file
    :param partitions: list containing aws partition addresses
    :param destination_dict: folder where results are saved
    """
    df = pd.DataFrame()
    for part_id in partitions:
        partial_df = pd.read_csv(destination_dict + part_id.local_output_file_name, index_col=0)
        df = df.append(partial_df, ignore_index=True)
        df.index = np.arange(len(df))
    return df


def check_completion(ssh_username, instance_id, ssh_key_file):
    """
    Checks for estimation estimation in partitions
    :param ssh_username: aws username
    :param instance_id: id of the aws instance
    :param ssh_key_file: full path to key file of aws_username
    :return: boolean, True if all partitions are finished
    """
    commands = ["grep -a 'finished' ./out"]
    commands_dict = remote_execute(user=ssh_username, instance_id=instance_id, key=ssh_key_file,
                                   shell_commands=commands, verbose=False)
    if str(commands_dict["grep -a 'finished' ./out"][0]).find('finished') != -1:
        return True
    else:
        return False


def main(estimation, df, ec2_instances, site_input_file, output_folder_location, ssh_key_file, aws_username,
         aws_instance_name, aws_region, aws_client, script_name, script_location, conda_environment, power_column_id,
         convert_to_ts, s3_location, n_files, file_label, fix_time_shifts, time_zone_correction, check_json,
         supplementary_file, data_source, gmt_offset):
    # number of partitions
    n_part = len(ec2_instances)
    total_size = np.sum(df['file_size'])
    # total_size = len(df)
    # size of partition
    part_size = np.ceil(total_size / n_part) * 0.8
    ii = 0
    jj = 0
    partitions = []

    for i in range(n_part):
        local_size = 0
        while local_size < part_size:
            local_size = np.sum(df.loc[ii:jj, 'file_size'])
            if i == n_part - 1:
                jj = len(df) - 1
                local_size = part_size + 1
            jj += 1
        # create partition
        part = get_config(est=estimation, part_id=i, ix_0=ii, ix_n=jj, n_part=n_part, ifl=site_input_file,
                          ofl=output_folder_location, ip_address=ec2_instances[i], skf=ssh_key_file, au=aws_username,
                          ain=aws_instance_name, ar=aws_region, ac=aws_client, script_name=script_name,
                          scripts_location=script_location, conda_env=conda_environment, pcid=power_column_id,
                          cts=convert_to_ts, s3l=s3_location, n_files=n_files, file_label=file_label,
                          fix_time_shifts=fix_time_shifts, time_zone_correction=time_zone_correction,
                          check_json=check_json, sup_file=supplementary_file, data_source=data_source,
                          gmt_offset=gmt_offset)
        # add partition to list
        partitions.append(part)
        create_partition(part)
        ii = jj + 1
        jj = ii

    completion = [False] * len(partitions)
    # check for completion
    while False in completion:
        print(' ')
        for part_ix, part_id in enumerate(partitions):
            if completion[part_ix] is False:
                ssh_key_file = part_id.ssh_key_file
                instance = part_id.public_ip_address
                ssh_username = part_id.aws_username
                new_value = check_completion(ssh_username, instance, ssh_key_file)
                part_id.process_completed = new_value
                completion[part_ix] = new_value
                if new_value is False:
                    status = 'running'
                else:
                    status = 'finished'
                print('partition' + ' ' + str(part_ix) + ':' + ' ' + status)

        time.sleep(10 * 60)
    # collect local result files
    get_remote_output_files(partitions, main_class.aws_username, main_class.global_output_directory)
    # combine results files
    results_df = combine_results(partitions, main_class.global_output_directory)
    # save consolidated results file
    results_df.to_csv(main_class.global_output_file)
    return


if __name__ == '__main__':
    """
    :param estimation: Estimation to be performed. Options are 'report', 'longitude', 'latitude', 'tilt_azimuth'
    :param input_site_file:  csv file containing list of sites to be evaluated. 'None' if no input file is provided.
    :param n_files: number of files to read. If 'all' all files in folder are read.
    :param s3_location: Absolute path to s3 location of files.
    :param file_label:  Repeating portion of data files label. If 'None', no file label is used. 
    :param power_column_label: Repeating portion of the power column label. 
    :param output_file: Absolute path to csv file containing report results.
    :param fix_time_shits: String, 'True' or 'False'. Specifies if time shifts are to be 
    fixed when running the pipeline.
    :param time_zone_correction: String, 'True' or 'False'. Specifies if the time zone correction is performed when 
    running the pipeline.
    :param check_json: String, 'True' or 'False'. Check json file for location information.
    :param convert_to_ts: String, 'True' or 'False'. Specifies if conversion to time series is performed when 
    running the pipeline.
    :param system_summary_file: Full path to csv file containing longitude and manual time shift flag for each system,
    None if no file
    provided. 
    :param gmt_offset: String. Single value of gmt offset to be used for all estimations. If None a list with individual
    gmt offsets needs to be provided.
    :param data_source: String. Input signal data source. Options are 's3' and 'cassandra'.
    :param script_to_execute: Full path to python script to be executed.
    :param conda environment: conda environment used to run script_to_execute.
    :param aws_instance_name: name of amazon web services instances used for running the study. All instances must have
    the same name
    """

    input_kwargs = sys.argv
    inputs_dict = get_commandline_inputs(input_kwargs)
    # The three input arguments below are required in addition to the input arguments required by the run scripts.
    # They are related to 'aws' partition handling.
    script_to_execute = str(sys.argv[-3])
    conda_environment = str(sys.argv[-2])
    aws_instance_name = str(sys.argv[-1])

    estimation = inputs_dict['estimation']
    input_site_file = inputs_dict['input_site_file']
    n_files = inputs_dict['n_files']
    s3_location = inputs_dict['s3_location']
    file_label = inputs_dict['file_label']
    power_column_label = inputs_dict['power_column_label']
    fix_time_shifts = inputs_dict['fix_time_shifts']
    time_zone_correction = inputs_dict['time_zone_correction']
    check_json = inputs_dict['check_json']
    convert_to_ts = inputs_dict['convert_to_ts']
    system_summary_file = inputs_dict['system_summary_file']
    gmt_offset = inputs_dict['gmt_offset']
    data_source = inputs_dict['data_source']

    # Default input variables
    if not input_site_file:
        build_input_file(s3_location)
        input_site_file = 's3://pv.insight.misc/report_files/generated_site_list.csv'

    aws_username = 'ubuntu'
    aws_region = 'us-west-1'
    aws_client = 'ec2'
    output_folder_location = '~/'
    global_output_directory = '~/results/'
    global_output_file = 'results.csv'
    pos = script_to_execute.rfind('/') + 1
    script_location = script_to_execute[:pos]
    script_name = script_to_execute.split('/')[-1]

    # aws licence file
    try:
        ssh_key_file = glob.glob("/Users/*/.aws/*.pem")[0]
    except:
        ssh_key_file = glob.glob("/home/*/.aws/*.pem")[0]

    # create main class
    main_class = get_config(est=estimation, ifl=input_site_file, ofl=output_folder_location, skf=ssh_key_file, au=aws_username,
                            ain=aws_instance_name, ar=aws_region, ac=aws_client, pcid=power_column_label,
                            gof=global_output_file, god=global_output_directory, cts=convert_to_ts,
                            s3l=s3_location, n_files=n_files, file_label=file_label,
                            time_zone_correction=time_zone_correction, check_json=check_json,
                            sup_file=system_summary_file, data_source=data_source, gmt_offset=gmt_offset)
    # collect aws instance addresses
    ec2_instances = get_address(aws_instance_name, aws_region, aws_client)
    # read input site file
    df = pd.read_csv(input_site_file, index_col=0)

    main(estimation, df, ec2_instances, input_site_file, output_folder_location, ssh_key_file, aws_username,
         aws_instance_name, aws_region, aws_client, script_name, script_location, conda_environment, power_column_label,
         convert_to_ts, s3_location, n_files, file_label, fix_time_shifts, time_zone_correction, check_json,
         system_summary_file, data_source, gmt_offset)



import unittest
import os
from pathlib import Path
import numpy as np
path = Path.cwd().parent.parent
os.chdir(path)
from pvsystemprofiler.algorithms.angle_of_incidence.lambda_functions import select_function
from pvsystemprofiler.algorithms.angle_of_incidence.curve_fitting import run_curve_fit

class TestFitAzimuth(unittest.TestCase):

    def test_fit_azimuth(self):
        # INPUTS

        filepath = Path(__file__).parent.parent
        # delta_f
        filepath = Path(__file__).parent.parent
        delta_f_file_path = filepath / "fixtures" / "tilt_azimuth" / "delta_f.csv"
        with open(delta_f_file_path) as file:
                delta_f = np.genfromtxt(file, delimiter=',')
        # omega_f
        filepath = Path(__file__).parent.parent
        omega_f_file_path = filepath / "fixtures" / "tilt_azimuth" / "omega_f.csv"
        with open(omega_f_file_path) as file:
                omega_f = np.genfromtxt(file, delimiter=',')
        # costheta_fit
        filepath = Path(__file__).parent.parent
        costheta_fit_file_path = filepath / "fixtures" / "tilt_azimuth" / "costheta_fit.csv"
        with open(costheta_fit_file_path) as file:
                costheta_fit = np.genfromtxt(file, delimiter=',')
        # boolean_filter
        filepath = Path(__file__).parent.parent
        boolean_filter_file_path = filepath / "fixtures" / "tilt_azimuth" / "boolean_filter.csv"
        with open(boolean_filter_file_path) as file:
                boolean_filter = np.genfromtxt(file, delimiter=',')
                boolean_filter = boolean_filter.astype(dtype=bool)
        # keys
        keys = ['tilt', 'azimuth']
        # init_values
        init_values = [30, 30]

        # Expected Tilt and azimuth output is generated in tests/fixtures/tilt_azimuth/tilt_azimuth_Estimation_data_creator.ipynb
        expected_output = 1.654457422429566
        func_customized, bounds = select_function(39.4856, None, None)     
        actual_output = run_curve_fit(func=func_customized, keys=['tilt', 'azimuth'], delta=delta_f, omega=omega_f, costheta=costheta_fit, boolean_filter=boolean_filter, init_values=[30, 30], fit_bounds=bounds)[1]
        np.testing.assert_almost_equal(actual_output, expected_output)

if __name__ == '__main__':
    unittest.main()



"""Omega, the hour angle is estimated as defined on p. 13 in:
       Duffie, John A., and William A. Beckman. Solar engineering of thermal
       processes. New York: Wiley, 1991."""

import numpy as np
from pvsystemprofiler.utilities.time_convert import clock_to_solar


def calculate_omega(data_sampling, num_days, lon, doy, gmt_offset):
    """
        :param data_sampling: daily data sampling.
        :param num_days: number of sampling days.
        :param lon: longitude in Degrees (float)
        :param doy: day of year (float or array)
        :param gmt_offset: local timezone offset in hours from UTC/GMT (float or int)
        :return: hour angle omega (float or array)
        """
    minutes_day = np.arange(0, 1440, data_sampling)
    minutes_doy = np.tile(minutes_day.reshape(-1, 1), (1, num_days))
    hours_doy_solar = clock_to_solar(minutes_doy, lon, doy, gmt_offset, eot='duffie')
    hours_doy_solar /= 60
    omega = 15 * (hours_doy_solar - 12)
    return omega



import numpy as np

"""
Calculates the angle incidence using the cos(theta) equation (1.6.2) in:
Duffie, John A., and William A. Beckman. Solar engineering of thermal processes. New York: Wiley, 1991.
"""


def calculate_costheta(func, delta, omega, lat=None, tilt=None, azim=None):
    """
    :param func: angle of incidence model function.
    :param delta: System's declination in Degrees (array).
    :param omega: System's hour angle in Degrees (array).
    :param lat: System's latitude in Degrees.
    :param tilt: System's tilt in Degrees.
    :param azim: System's azimuth in Degrees.
    :return: angle of incidence array
    """
    x = np.array([np.deg2rad(delta), np.deg2rad(omega)])
    phi = np.deg2rad(lat)
    beta = np.deg2rad(tilt)
    gamma = np.deg2rad(azim)
    costheta = func(x, phi, beta, gamma)
    return costheta



import sys


def progress(count, total, status='', bar_length=60):
    """
    Python command line progress bar in less than 10 lines of code. · GitHub
    https://gist.github.com/vladignatyev/06860ec2040cb497f0f3
    :param count: the current count, int
    :param total: to total count, int
    :param status: a message to display
    :return:
    """
    bar_len = bar_length
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)

    sys.stdout.write('[%s] %s%s ...%s\r' % (bar, percents, '%', status))
    sys.stdout.flush()



from pvsystemprofiler.algorithms.longitude.calculation import calculate_longitude
from pvsystemprofiler.algorithms.longitude.fitting import fit_longitude


def estimate_longitude(estimator, eot, solarnoon, days, gmt_offset):
    if estimator == 'calculated':
        return calculate_longitude(eot, solarnoon, days, gmt_offset)
    else:
        loss = estimator.split('_')[-1]
        return fit_longitude(eot, solarnoon, days, gmt_offset, loss=loss)



import numpy as np
""" Longitude Direct Calculation Module
This module contains the function for the direct calculation of system
longitude based on estimated local solar noon and timezone offset from UTC.
The same exact equation is used for "Hadghdadi" and "Duffie" approaches. See
`pvsystemprofiler.utilities.equation_of_time` for equation of time (EoT)
calculations.
"""


def calculate_longitude(eot, solarnoon, days, gmt_offset):
    sn = 60 * solarnoon[days]  # convert hours to minutes
    eot_days = eot[days]
    estimates = calc_lon(sn, eot_days, gmt_offset)
    return np.nanmedian(estimates)


def calc_lon(solar_noon, eot, gmt_offset):
    """
    Using the definition of solar time and standard time, given in equation
    (1.5.2) in [1]. This is the standard equation governing the relationship
    between apparent time and the equation of time. Note that 720 is the
    number of minutes in twelve hours, so it corresponds with "standard noon"
    or 12:00PM, given in minutes since midnight.

    [1] Duffie, John A., and William A. Beckman. Solar engineering of thermal
        processes. New York: Wiley, 1991.

    :param solar_noon: time of solar noon in minutes on a given day
    :param eot: equation of time on a given day
    :param gmt_offset: local timezone offset in hours from UTC/GMT
    :return: the longitude for the given values
    """
    sn = solar_noon
    tc = 720 - sn
    lon = (tc - eot) / 4 + 15 * gmt_offset
    return lon



import numpy as np


def delta_cooper(day_of_year, daily_meas):
    """"
    Declination delta is estimated from equation (1.6.1a) in:
    Duffie, John A., and William A. Beckman. Solar engineering of thermal
    processes. New York: Wiley, 1991.
    """
    delta_1 = 23.45 * np.sin(np.deg2rad(360 * (284 + day_of_year) / 365))
    delta = np.tile(delta_1, (daily_meas, 1))
    return delta


def delta_spencer(day_of_year, daily_meas):
    """"
    Declination delta is estimated from equation (1.6.1b) in:
    Duffie, John A., and William A. Beckman. Solar engineering of thermal
    processes. New York: Wiley, 1991.
    """
    b = np.deg2rad((day_of_year - 1) * 360 / 365)
    delta_1 = (180 / np.pi) * (0.006918 - 0.399912 * np.cos(b) + 0.070257 * np.sin(b) - 0.006758 *
                               np.cos(2 * b) + 0.000907 * np.sin(2 * b) - 0.002697 * np.cos(3 * b) +
                               0.00148 * np.sin(3 * b))
    delta = np.tile(delta_1, (daily_meas, 1))
    return delta



class ConfigPartitions:
    def __init__(self, est=None, part_id=None, ix_0=None, ix_n=None, n_part=None, ifl=None, ofl=None, ip_address=None,
                 skf=None, au=None, ain=None, ar=None, ac=None, script_name=None, scripts_location=None, conda_env=None,
                 pcid=None, gof=None, god=None, cts=None, s3l=None, n_files=None, file_label=None, fix_time_shifts=None,
                 time_zone_correction=None, check_json=None, sup_file=None, ds=None, gmt=None):
        self.estimation = est
        self.input_file_location = ifl
        self.ssh_key_file = skf
        self.aws_username = au
        self.aws_instance_name = ain
        self.aws_region = ar
        self.aws_client = ac
        self.supplementary_file = sup_file
        if ix_0 is not None and ix_n is not None:
            self.part_id = part_id
            self.env_name = 'partition_' + str(part_id)
            self.n_part = n_part
            self.local_working_folder_location = ofl
            self.local_working_folder = self.local_working_folder_location + 'estimation_part_{}_of_{}/'.format(
                str(part_id + 1), str(n_part))
            self.local_input_file = self.local_working_folder + 'input_part_{}_of_{}.csv'.format(str(part_id + 1),
                                                                                                 str(n_part))
            self.local_output_file_name = 'results_part_{}_of_{}.csv'.format(str(part_id + 1), str(n_part))
            self.local_output_file = self.local_working_folder + self.local_output_file_name
            self.ix_0 = ix_0
            self.ix_n = ix_n
            self.scripts_location = scripts_location
            self.script_name = self.scripts_location + script_name
            self.conda_environment = conda_env
            self.power_column_id = pcid
            self.convert_to_ts = cts
            self.public_ip_address = ip_address
            self.process_completed = False
            self.s3_location = s3l
            self.n_files = n_files
            self.file_label = file_label
            self.fix_time_shifts = fix_time_shifts
            self.time_zone_correction = time_zone_correction
            self.check_json = check_json
            self.data_source = ds
            self.gmt_offset = gmt
        else:
            self.global_output_directory = god
            self.global_output_file = god + gof


def get_config(est=None, part_id=None, ix_0=None, ix_n=None, n_part=None, ifl=None, ofl=None, ip_address=None, skf=None,
               au=None, ain=None, ar=None, ac=None, script_name=None, scripts_location=None, conda_env=None, pcid=None,
               gof=None, god=None, cts=None, s3l=None, n_files=None, file_label=None, fix_time_shifts=None,
               time_zone_correction=None, check_json=None, sup_file=None, data_source=None, gmt_offset=None):
    if ix_0 is not None and ix_n is not None:
        return ConfigPartitions(est=est, part_id=part_id, ix_0=ix_0, ix_n=ix_n, n_part=n_part, ifl=ifl, ofl=ofl,
                                ip_address=ip_address, skf=skf, au=au, ain=ain, ar=ar, ac=ac, script_name=script_name,
                                scripts_location=scripts_location, conda_env=conda_env, pcid=pcid, cts=cts, s3l=s3l,
                                n_files=n_files, file_label=file_label, fix_time_shifts=fix_time_shifts,
                                time_zone_correction=time_zone_correction, check_json=check_json, sup_file=sup_file,
                                ds=data_source, gmt=gmt_offset)
    else:
        return ConfigPartitions(est=est, ifl=ifl, skf=skf, au=au, ain=ain, ar=ar, ac=ac, gof=gof, god=god)



import unittest
import os
from pathlib import Path
import numpy as np
path = Path.cwd().parent.parent
os.chdir(path)
from pvsystemprofiler.algorithms.latitude.estimation import estimate_latitude


class TestEstimateLatitude(unittest.TestCase):

    def test_estimate_latitude(self):
        # INPUTS
        filepath = Path(__file__).parent.parent
        # hours daylight
        hours_daylight_file_path = filepath / "fixtures" / "latitude" / "hours_daylight.csv"
        with open(hours_daylight_file_path) as file:
                hours_daylight = np.genfromtxt(file, delimiter=',')
        # Delta
        delta_file_path = filepath / "fixtures" / "latitude" / "delta.csv"
        with open(delta_file_path) as file:
                delta = np.genfromtxt(file, delimiter=',')

        # Expected Latitude Output is generated in tests/fixtures/latitude/latitude_test_data_creator.ipynb
        expected_output =  38.58601372121755

        actual_output = estimate_latitude(hours_daylight, delta)
        np.testing.assert_almost_equal(actual_output, expected_output, decimal=1)


if __name__ == '__main__':
    unittest.main()



from pvsystemprofiler.algorithms.latitude.calculation import calculate_latitude


def estimate_latitude(hours_daylight, delta):
    latitude_estimate = calculate_latitude(hours_daylight, delta)
    return latitude_estimate




import sys
from pathlib import Path

# TODO: remove pth.append after package is deployed
filepath = Path(__file__).resolve().parents[1]
sys.path.append(str(filepath))
from pvsystemprofiler.scripts.modules.script_functions import remote_execute


def create_partition(partition):
    """
    Creates remote partition from `run_partition_script`
    :param partition: object containing information about each individual partition
    """
    start_index = partition.ix_0
    end_index = partition.ix_n
    global_input_file = partition.input_file_location
    local_input_file = partition.local_input_file
    local_working_folder = partition.local_working_folder
    local_output_file = partition.local_output_file
    script_name = partition.script_name
    estimation = partition.estimation
    scripts_location = partition.scripts_location
    local_script = scripts_location + 'modules/local_partition_script.py'
    conda_env = partition.conda_environment
    instance = partition.public_ip_address
    ssh_username = partition.aws_username
    ssh_key_file = partition.ssh_key_file
    power_column_id = partition.power_column_id
    convert_to_ts = partition.convert_to_ts
    s3_location = partition.s3_location
    n_files = partition.n_files
    file_label = partition.file_label
    fix_time_shifts = partition.fix_time_shifts
    time_zone_correction = partition.time_zone_correction
    check_json = partition.check_json
    supplementary_file = partition.supplementary_file
    data_type = partition.data_source
    gmt_offset = partition.gmt_offset

    # prepare python command to run local partition
    # extract conda installation folder from local .bashrc
    grep_conda = "grep '__conda_setup=' .bashrc"
    commands = [grep_conda]
    output = remote_execute(ssh_username, instance, ssh_key_file, commands)
    conda_location = output[grep_conda][0]
    conda_location = conda_location.decode('utf-8')
    i = conda_location.find("('")
    j = conda_location.find("bin", i + 2)
    conda_location = conda_location[i + 2: j]
    python_command = conda_location + 'envs/' + conda_env + '/bin/python'

    # delete existing remote partitions
    commands = ['rm estimation* -rf']
    output = remote_execute(ssh_username, instance, ssh_key_file, commands)

    # check for previously created remote folders
    commands = ['ls' + ' ' + local_working_folder]
    output = remote_execute(ssh_username, instance, ssh_key_file, commands)

    # create remote partition if does not exist
    if str(output[commands[0]][1]).find('No such file or directory') != -1:
        # prepare commands to create partition
        commands = ['rm estimation* -rf',
                    'mkdir' + ' ' + local_working_folder,
                    python_command + ' ' + local_script + ' '
                    + str(start_index) + ' '
                    + str(end_index) + ' '
                    + script_name + ' '
                    + estimation + ' '
                    + global_input_file + ' '
                    + local_working_folder + ' '
                    + local_input_file + ' '
                    + local_output_file + ' '
                    + power_column_id + ' '
                    + str(convert_to_ts) + ' '
                    + s3_location + ' '
                    + n_files + ' '
                    + str(file_label) + ' '
                    + str(fix_time_shifts) + ' '
                    + str(time_zone_correction) + ' '
                    + str(check_json) + ' '
                    + str(gmt_offset) + ' '
                    + data_type + ' '
                    + supplementary_file + ' '
                    + python_command
                    ]
    else:
        # if remote partition exist from previous run, resume run
        commands = [local_working_folder + 'run_local_partition.sh']
    # execute remote partition script
    remote_execute(ssh_username, instance, ssh_key_file, commands)



"""
This module contains a class for estimating longitude, latitude, tilt and azimuth from an input signal. This code
accepts an input signal data in the form of a `solar-data-tools` `DataHandler` object, which is used to standardize
and pre-process the data. The provided class will then estimate  longitude, latitude, tilt and azimuth using the
 'estimate_longitude', 'estimate_latitude' and 'estimate_orientation' methods, respectively. Alternatively, all four
 parameters can be estimated at once using the 'estimate_all' method.
"""
# Standard Imports
import numpy as np
# Solar Data Tools Imports
from solardatatools.solar_noon import energy_com, avg_sunrise_sunset
# Module Imports
from pvsystemprofiler.utilities.equation_of_time import eot_da_rosa, eot_duffie
from solardatatools.algorithms import SunriseSunset
from pvsystemprofiler.algorithms.latitude.hours_daylight import calculate_hours_daylight
from pvsystemprofiler.utilities.hour_angle_equation import calculate_omega
from pvsystemprofiler.utilities.declination_equation import delta_cooper
from pvsystemprofiler.algorithms.angle_of_incidence.curve_fitting import run_curve_fit
from pvsystemprofiler.algorithms.performance_model_estimation import find_fit_costheta
from pvsystemprofiler.algorithms.angle_of_incidence.lambda_functions import select_function
from pvsystemprofiler.algorithms.angle_of_incidence.dynamic_value_functions import determine_keys
from pvsystemprofiler.algorithms.angle_of_incidence.dynamic_value_functions import select_init_values
from pvsystemprofiler.algorithms.tilt_azimuth.daytime_threshold_quantile import filter_data
from pvsystemprofiler.utilities.tools import random_initial_values
from pvsystemprofiler.algorithms.longitude.estimation import estimate_longitude
from pvsystemprofiler.algorithms.latitude.estimation import estimate_latitude


class ConfigurationEstimator():
    def __init__(self, data_handler, gmt_offset, day_selection_method='all', solar_noon_method='optimized_estimates',
                 daylight_method='optimized_estimates', data_matrix='filled', daytime_threshold=None):
        if not data_handler._ran_pipeline:
            data_handler.run_pipeline()
        self.data_handler = data_handler
        if data_matrix == 'raw':
            self.data_matrix = data_handler.raw_data_matrix
        elif data_matrix == 'filled':
            self.data_matrix = data_handler.filled_data_matrix
        # Parameters to be estimated
        self.longitude = None
        self.latitude = None
        self.tilt = None
        self.azimuth = None
        # Attributes used for all calculations
        self.gmt_offset = gmt_offset
        self.hours_daylight = None
        self.daily_meas = self.data_handler.filled_data_matrix.shape[0]
        self.daytime_threshold = None
        self.day_interval = None
        self.x1 = None
        self.x2 = None
        self.data_sampling = self.data_handler.data_sampling
        self.num_days = self.data_handler.num_days
        self.day_of_year = self.data_handler.day_index.dayofyear
        self.eot_duffie = eot_duffie(self.day_of_year)
        self.eot_da_rosa = eot_da_rosa(self.day_of_year)
        self.delta = delta_cooper(self.day_of_year, self.daily_meas)
        self.omega = None
        if day_selection_method == 'all':
            self.days = self.data_handler.daily_flags.no_errors
        elif day_selection_method == 'clear':
            self.days = self.days.daily_flags.clear
        elif day_selection_method == 'cloudy':
            self.days = self.data_handler.daily_flags.cloudy

        if solar_noon_method == 'optimized_estimates' or daylight_method == 'optimized_estimates':
            ss = SunriseSunset()

        if solar_noon_method == 'rise_set_average':
            self.solarnoon = avg_sunrise_sunset(self.data_matrix)
        elif solar_noon_method == 'energy_com':
            self.solarnoon = energy_com(self.data_matrix)
        if solar_noon_method == 'optimized_estimates':
            ss.run_optimizer(data=self.data_matrix)
            self.solarnoon = np.nanmean([ss.sunrise_estimates, ss.sunset_estimates], axis=0)

        if daylight_method in ('sunrise-sunset', 'sunrise sunset'):
            self.hours_daylight = calculate_hours_daylight(self.data_matrix, daytime_threshold)
        elif daylight_method == 'optimized_estimates':
            self.hours_daylight = ss.sunset_estimates - ss.sunrise_estimates

    def estimate_longitude(self, estimator='fit_l1', eot_calculation='duffie'):
        """
        :param estimator: 'calculated', 'fit_l1', 'fit_l2' or 'fit_huber'
        :param eot_calculation: 'rise_set_average', 'energy_com', 'optimized_estimates',
        or 'optimized_measurements'
        :return: None
        """
        if eot_calculation in ('duffie', 'd', 'duf'):
            eot = self.eot_duffie
        elif eot_calculation in ('da_rosa', 'dr', 'rosa'):
            eot = self.eot_da_rosa
        self.longitude = estimate_longitude(estimator, eot, self.solarnoon, self.days, self.gmt_offset)
        return

    def estimate_latitude(self):
        hours_daylight_filtered, delta_filtered = self._prepare_lat_input_data()
        self.latitude = estimate_latitude(hours_daylight_filtered, delta_filtered)
        return

    def _prepare_lat_input_data(self):
        if np.any(np.isnan(self.hours_daylight)):
            hours_mask = np.isnan(self.hours_daylight)
            full_mask = ~hours_mask & self.days
            hours_daylight_filtered = self.hours_daylight[full_mask]
            delta = self.delta[:, full_mask]
        else:
            hours_daylight_filtered = self.hours_daylight[self.days]
            delta = self.delta[:, self.days]
        return hours_daylight_filtered, delta

    # estimate tilt and azimuth with or without longitude and latitude input values
    def estimate_orientation(self, longitude=None, latitude=None, tilt=None, azimuth=None, day_interval=None, x1=0.9,
                             x2=0.9):
        """
        Estimates tilt and azimuth. The intended use is to estimate tilt and azimuth given longitude and latitude.
        However, the algorithm will estimate any of longitude, latitude, tilt and azimuth depending on the input values.
        If a parameter is not provided as an input value, it will be estimated. Any other use than the stated above as
        the intended use was found to yield inaccurate results.

        :param longitude: optional. Longitude value to be used in parameter estimation.
        :param latitude: optional. Latitude value to be used in parameter estimation.
        :param tilt: optional. Tilt value to be used in parameter estimation.
        :param azimuth: optional. Azimuth value to be used in parameter estimation.
        :param day_interval: 'all', 'clear' or 'cloudy'.
        :param x1: cvx parameter. Factor used in signal decomposition for estimation of daytime threshold.
        :param x2: Quantile of data used in estimation of daytime threshold.
        :return: None
        """

        if longitude is None:
            est_lon = ConfigurationEstimator(self.data_handler, self.gmt_offset)
            est_lon.estimate_longitude()
            self.longitude = est_lon.longitude
        else:
            self.longitude = longitude
        if latitude is None:
            est_lat = ConfigurationEstimator(self.data_handler, self.gmt_offset)
            est_lat.estimate_latitude()
            self.latitude = est_lat.latitude
        else:
            self.latitude = latitude

        self.tilt = tilt
        self.azimuth = azimuth
        self.day_interval = day_interval
        self.x1 = x1
        self.x2 = x2
        dh = self.data_handler
        self.data_matrix = dh.filled_data_matrix
        self.num_days = dh.num_days
        self.omega = calculate_omega(self.data_sampling, self.num_days, self.longitude, self.day_of_year,
                                     self.gmt_offset)

        self.tilt, self.azimuth = self._cal_orientation_helper()

    def estimate_all(self, day_interval=None, x1=0.9, x2=0.9):
        """
        Estimate latitude, longitude, tilt and azimuth all at once.
        :param day_interval: 'all', 'clear', 'cloudy'.
        :param x1: cvx parameter. Factor used in signal decomposition for estimation of daytime threshold.
        :param x2: Quantile of data used in estimation of daytime threshold.
        """

        self.tilt = None
        self.azimuth = None
        self.day_interval = day_interval
        self.x1 = x1
        self.x2 = x2
        dh = self.data_handler
        self.data_matrix = dh.filled_data_matrix
        self.days = dh.daily_flags.clear
        self.num_days = dh.num_days
        self.delta = delta_cooper(self.day_of_year, self.daily_meas)
        est_lon = ConfigurationEstimator(self.data_handler, self.gmt_offset)
        est_lon.estimate_longitude()
        self.longitude = est_lon.longitude
        est_lat = ConfigurationEstimator(self.data_handler, self.gmt_offset)
        est_lat.estimate_latitude()
        self.latitude = est_lat.latitude
        self.omega = calculate_omega(self.data_sampling, self.num_days, self.longitude, self.day_of_year,
                                     self.gmt_offset)

        self.tilt, self.azimuth = self._cal_orientation_helper()

    def _cal_orientation_helper(self):
        if self.day_interval is not None:
            day_range = (self.day_of_year > self.day_interval[0]) & (self.day_of_year < self.day_interval[1])
        else:
            day_range = np.ones(self.day_of_year.shape, dtype=bool)

        scale_factor_costheta, costheta_fit = find_fit_costheta(self.data_matrix, self.days)

        boolean_filter = filter_data(self.data_matrix, self.daytime_threshold, self.x1, self.x2)

        boolean_filter = boolean_filter * self.days * day_range

        delta_f = self.delta[boolean_filter]
        omega_f = self.omega[boolean_filter]
        if ~np.any(boolean_filter):
            print('No data made it through filters')

        lat_initial, tilt_initial, azim_initial = random_initial_values(1)

        func_customized, bounds = select_function(self.latitude, self.tilt, self.azimuth)
        dict_keys = determine_keys(latitude=self.latitude, tilt=self.tilt, azimuth=self.azimuth)

        init_values_dict = {'latitude': lat_initial[0], 'tilt': tilt_initial[0], 'azimuth': azim_initial[0]}
        init_values, ivr = select_init_values(init_values_dict, dict_keys)

        estimates = run_curve_fit(func=func_customized, keys=dict_keys, delta=delta_f, omega=omega_f,
                                  costheta=costheta_fit, boolean_filter=boolean_filter, init_values=init_values,
                                  fit_bounds=bounds)

        for i, estimate in enumerate(dict_keys):
            if estimate == 'latitude_estimate':
                lat_estimate = estimates[i]
            if estimate == 'tilt_estimate':
                tilt_estimate = estimates[i]
            if estimate == 'azimuth_estimate':
                azimuth_estimate = estimates[i]

        if 'tilt_estimate' not in dict_keys:
            tilt_estimate = None
        if 'azimuth_estimate' not in dict_keys:
            azimuth_estimate = None
        return tilt_estimate, azimuth_estimate



import unittest
import os
from pathlib import Path
import numpy as np
path = Path.cwd().parent.parent
os.chdir(path)
from pvsystemprofiler.algorithms.longitude.calculation import calculate_longitude


class TestCalculateLongitude(unittest.TestCase):

    def test_calculate_longitude(self):
        # INPUTS
        filepath = Path(__file__).parent.parent
        # eot_duffie
        eot_duffie_file_path = filepath / "fixtures" / "longitude" / "eot_duffie_output.csv"
        with open(eot_duffie_file_path) as file:
                eot_duffie = np.genfromtxt(file, delimiter=',')
        # solarnoon
        solarnoon_file_path = filepath / "fixtures" / "longitude" / "solarnoon.csv"
        with open(solarnoon_file_path) as file:
                solarnoon = np.genfromtxt(file, delimiter=',')
        # days
        days_file_path = filepath / "fixtures" / "longitude" / "days.csv"
        with open(days_file_path) as file:
                days = np.genfromtxt(file, delimiter=',')
                days = days.astype(dtype=bool)
        # gmt_offset
        gmt_offset = -5
        # loss
        loss = 'l2'

        # Expected Longitude Output is generated in tests/fixtures/longitude/longitude_fitting_and_calculation_test_data_creator.ipynb
        expected_output =  -77.10636729272031

        actual_output = calculate_longitude(eot_duffie, solarnoon, days, gmt_offset)
        np.testing.assert_almost_equal(actual_output, expected_output, decimal=1)


if __name__ == '__main__':
    unittest.main()



""" Longitude Direct Calculation Module
This module contains the function for the direct calculation of a system's
latitude based on daylight hours and the declination angle.
"""
import numpy as np


def calculate_latitude(hours_daylight, delta):
    estimates = calc_lat(hours_daylight, delta)
    return np.nanmedian(estimates)


def calc_lat(hours_daylight, delta):
    """
    Latitude is estimated from equation (1.6.11) in:
    Duffie, John A., and William A. Beckman. Solar engineering of thermal
    processes. New York: Wiley, 1991.

    :param hours_daylight: daylight hours as calculated by calculate_hours_daylight or calculate_hours_daylight_raw
    :param delta: declination as calculated from declination_equations in Degrees.
    :return: the latitude for the given values
    """
    lat = np.degrees(np.arctan(- np.cos(np.radians(15 / 2 * hours_daylight)) / (np.tan(np.deg2rad(delta[0])))))
    return lat



import unittest
import os
from pathlib import Path
import numpy as np
path = Path.cwd().parent.parent
os.chdir(path)
from pvsystemprofiler.utilities.equation_of_time import eot_da_rosa, eot_duffie


class TestEquationOfTime(unittest.TestCase):

    # importing input for both eot tests
    filepath = Path(__file__).parent.parent
    input_data_file_path = filepath / 'fixtures' / 'longitude' / 'eot_input.csv'
    with open(input_data_file_path) as file:
            input_data = np.genfromtxt(file, delimiter=',')


    def test_eot_duffie(self):

        expected_data_file_path = self.filepath / 'fixtures' / 'longitude' / 'eot_duffie_output.csv'
        with open(expected_data_file_path) as file:
            expected_output = np.genfromtxt(file, delimiter=',')

        actual_output = eot_duffie(self.input_data)
        np.testing.assert_array_almost_equal(actual_output, expected_output)


    def test_eot_da_rosa(self):

        expected_data_file_path = self.filepath / 'fixtures' / 'longitude' / 'eot_da_rosa_output.csv'
        with open(expected_data_file_path) as file:
            expected_output = np.genfromtxt(file, delimiter=',')

        actual_output = eot_da_rosa(self.input_data)
        np.testing.assert_array_almost_equal(actual_output, expected_output)


if __name__ == '__main__':
    unittest.main()



""" Tilt and Azimuth Study Module
This module contains a class for conducting a study to estimating Tilt and Azimuth from an input signal. This code
accepts an input signal data in the form of a `solar-data-tools` `DataHandler` object, which is used to standardize
and pre-process the data. The provided class will then estimate the Tilt and Azimuth of the site that produced the data,
using the `run` method. Tilt and Azimuth are estimated via numerical fit using equation (1.6.2) in:
Duffie, John A., and William A. Beckman. Solar engineering of thermal processes. New York: Wiley, 1991.

The following configurations can be run:
 - Day range: 'full_year' or customized day range
 - Declination equation: 'cooper', 'spencer'.
 """
import numpy as np
import pandas as pd
from pvsystemprofiler.utilities.hour_angle_equation import calculate_omega
from pvsystemprofiler.utilities.declination_equation import delta_spencer
from pvsystemprofiler.utilities.declination_equation import delta_cooper
from pvsystemprofiler.algorithms.angle_of_incidence.curve_fitting import run_curve_fit
from pvsystemprofiler.algorithms.angle_of_incidence.calculation import calculate_costheta
from pvsystemprofiler.algorithms.performance_model_estimation import find_fit_costheta
from pvsystemprofiler.algorithms.angle_of_incidence.lambda_functions import select_function
from pvsystemprofiler.utilities.angle_of_incidence_function import func_costheta
from pvsystemprofiler.algorithms.angle_of_incidence.dynamic_value_functions import determine_keys
from pvsystemprofiler.algorithms.angle_of_incidence.dynamic_value_functions import select_init_values
from pvsystemprofiler.utilities.tools import random_initial_values
from pvsystemprofiler.algorithms.tilt_azimuth.daytime_threshold_quantile import filter_data


class TiltAzimuthStudy():
    def __init__(self, data_handler, day_range='full_year', init_values=None, nrandom_init_values=None,
                 daytime_threshold=None, lon_input=None, lat_input=None, tilt_input=None,
                 azimuth_input=None, lat_true_value=None, tilt_true_value=None, azimuth_true_value=None,
                 gmt_offset=-8, cvx_parameter=None, threshold_quantile=None):
        """
        :param data_handler: `DataHandler` class instance loaded with a solar power data set.
        :param day_range: (optional) the desired day range to run the study. A list of the form
                              [first day, last day].
        :param init_values: (optional) Latitude, Tilt and Azimuth guess values list for numerical fit. A list of the
                form [[latitude_1,.., latitude_n], [tilt_1,.., tilt_n], [azimuth_1,.., azimuth_n]]. Default value is 10.
                (Degrees).
        :param nrandom_init_values: (optional) number of random initial values to be generated.
        :param daytime_threshold: (optional) daytime threshold
        :param lon_input: longitude estimate as obtained from the Longitude Study module in Degrees.
        :param lat_input: latitude value in Degrees.
        :param tilt_input: tilt value in Degrees.
        :param azimuth_input: azimuth value in Degrees.
        :param lat_true_value: (optional) ground truth value for the system's Latitude in Degrees.
        :param tilt_true_value: (optional) ground truth value for the system's Tilt in Degrees.
        :param azimuth_true_value: (optional) ground truth value for the system's Azimuth in Degrees Degrees.
        :param gmt_offset: The offset in hours between the local timezone and GMT/UTC.
        :param cvx_parameter: (optional). Factor used in signal decomposition for estimation of daytime threshold.
        :param threshold_quantile: (optional). Quantile of data used in estimation of daytime threshold.
        """

        self.data_handler = data_handler
        # Choose day range from season dictionary
        if day_range is None:
            self.day_range_dict = {}
            self.day_range_dict = {'summer': [171, 265], 'no_winter': [79, 355], 'spring': [79, 171],
                                   'winter': [355, 79], 'winter_spring': [355, 171], 'full_year': None}

        elif day_range is 'full_year':
            self.day_range_dict = {'full_year': None}
        elif day_range is 'manual':
            self.day_range_dict = {'manual': day_range}

        self.data_matrix = self.data_handler.filled_data_matrix
        if not data_handler._ran_pipeline:
            print('Running DataHandler preprocessing pipeline with defaults')
            self.data_handler.run_pipeline()
        # initial values
        self.init_values = init_values
        self.nrandom = nrandom_init_values
        # inputs
        self.lon_input = lon_input
        self.lat_input = lat_input
        self.tilt_input = tilt_input
        self.azimuth_input = azimuth_input
        # true values
        self.lat_true_value = lat_true_value
        self.tilt_true_value = tilt_true_value
        self.azimuth_true_value = azimuth_true_value
        self.gmt_offset = gmt_offset
        # thresholds
        self.daytime_threshold = daytime_threshold
        self.daytime_threshold_fit = None
        if cvx_parameter is None:
            self.threshold_x1 = [0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]
        else:
            self.threshold_x1 = np.atleast_1d(cvx_parameter)
        if threshold_quantile is None:
            self.threshold_x2 = [0.5, 0.55, 0.6, 0.65, 0.7, 0.75, 0.8, 0.85, 0.9, 0.95]
        else:
            self.threshold_x2 = np.atleast_1d(threshold_quantile)
        # data specific variables
        self.day_of_year = self.data_handler.day_index.dayofyear
        self.num_days = self.data_handler.num_days
        self.daily_meas = self.data_handler.filled_data_matrix.shape[0]
        self.data_sampling = self.data_handler.data_sampling
        self.clear_index = data_handler.daily_flags.clear
        # angle of incidence parameters
        self.delta_cooper = None
        self.delta_spencer = None
        self.omega = None
        self.scale_factor_costheta = None
        self.costheta_estimated = None
        self.costheta_ground_truth = None
        self.costheta_fit = None
        # other
        self.results = None

    def run(self, delta_method=('cooper', 'spencer')):
        """
        Run a study with the given configuration of options. Defaults to
        running all available options. Any kwarg can be constrained by
        providing a subset of acceptable keys. For example the default keys
        for the declination method estimator kwarg are:

        ('cooper', 'spencer')

        Additionally, any of the following would be acceptable for this kwarg:

        ('cooper')
        'cooper'
        'spencer'

        This method sets the `results` attribute to be a pandas data frame
        containing the results of the study.

        :param delta_method: 'cooper', 'spencer'.
        :return: None.
        """

        delta_method = np.atleast_1d(delta_method)
        # calculate hour angle
        self.omega = calculate_omega(self.data_sampling, self.num_days, self.lon_input, self.day_of_year,
                                     self.gmt_offset)
        # fit daily signal of cos theta
        self.scale_factor_costheta, self.costheta_fit = find_fit_costheta(self.data_matrix, self.clear_index)
        # estimate declination angles
        self.delta_cooper = delta_cooper(self.day_of_year, self.daily_meas)
        self.delta_spencer = delta_spencer(self.day_of_year, self.daily_meas)
        # initialize parameters
        if self.init_values is not None:
            lat_initial = self.init_values[0]
            tilt_initial = self.init_values[1]
            azim_initial = self.init_values[2]
        else:
            if self.nrandom is None:
                lat_initial = [10]
                tilt_initial = [10]
                azim_initial = [10]
            else:
                lat_initial, tilt_initial, azim_initial = random_initial_values(self.nrandom)

        counter = 0
        self.create_results_table()
        for x1 in self.threshold_x1:
            # first quantile value in signal decomposition algorithm
            for x2 in self.threshold_x2:
                # second quantile value in signal decomposition algorithm
                filtered_data = filter_data(self.data_matrix, self.daytime_threshold, x1, x2)
                for delta_id in delta_method:
                    # declination angle
                    if delta_id in ('Cooper', 'cooper'):
                        delta = self.delta_cooper
                    if delta_id in ('Spencer', 'spencer'):
                        delta = self.delta_spencer
                    for day_range_id in self.day_range_dict:
                        # day range
                        day_interval = self.day_range_dict[day_range_id]
                        boolean_filter = self.get_day_range(filtered_data, day_interval)
                        delta_f = delta[boolean_filter]
                        omega_f = self.omega[boolean_filter]
                        if ~np.any(boolean_filter):
                            print('No data made it through filters')
                        # choose function and unknowns based on provided inputs
                        # choose range for each unknown
                        func_customized, bounds = select_function(self.lat_input, self.tilt_input,
                                                                  self.azimuth_input)
                        # create dictionary keys for unknowns. If an input value for lat, tilt or azimuth is not
                        # provided, it will be tagged as an unknown
                        dict_keys = determine_keys(latitude=self.lat_input, tilt=self.tilt_input,
                                                   azimuth=self.azimuth_input)
                        nvalues = len(lat_initial)

                        for init_val_ix in np.arange(nvalues):
                            # loop over initial values
                            init_values_dict = {'latitude': lat_initial[init_val_ix], 'tilt': tilt_initial[init_val_ix],
                                                'azimuth': azim_initial[init_val_ix]}
                            init_values, ivr = select_init_values(init_values_dict, dict_keys)
                            try:
                                # estimate latitude and/or tilt and/or azimuth. If parameter is in keys, it will be
                                # estimated
                                estimates = run_curve_fit(func=func_customized, keys=dict_keys, delta=delta_f,
                                                          omega=omega_f, costheta=self.costheta_fit,
                                                          boolean_filter=boolean_filter, init_values=init_values,
                                                          fit_bounds=bounds)
                            except RuntimeError:
                                input_array = np.array([self.lat_input, self.tilt_input, self.azimuth_input])
                                estimates = np.full(np.sum(input_array == None), np.nan)
                            # create dictionary with dict_keys and estimates
                            estimates_dict = dict(zip(dict_keys, estimates))
                            # dynamic results dataFrame based on provided inputs
                            lat = estimates_dict[
                                'latitude_estimate'] if 'latitude_estimate' in estimates_dict else self.lat_input
                            tilt = estimates_dict['tilt_estimate'] if 'tilt_estimate' in estimates_dict else \
                                self.tilt_input
                            azim = estimates_dict[
                                'azimuth_estimate'] if 'azimuth_estimate' in estimates_dict else self.azimuth_input

                            self.costheta_estimated = calculate_costheta(func=func_costheta, delta=delta,
                                                                         omega=self.omega, lat=lat, tilt=tilt,
                                                                         azim=azim)
                            # calculate cos theta from analytical equation in case ground truth values are provided
                            if None not in (self.lat_true_value, self.tilt_true_value, self.azimuth_true_value):
                                self.costheta_ground_truth = calculate_costheta(func=func_costheta, delta=delta,
                                                                                omega=self.omega,
                                                                                lat=self.lat_true_value,
                                                                                tilt=self.tilt_true_value,
                                                                                azim=self.azimuth_true_value)

                            self.results.loc[counter] = [day_range_id, delta_id, x1, x2] + ivr + list(estimates)
                            counter += 1

        if self.lat_true_value is not None and self.lat_input is None:
            self.results['latitude residual'] = self.lat_true_value - self.results['latitude']
        if self.tilt_true_value is not None and self.tilt_input is None:
            self.results['tilt residual'] = self.tilt_true_value - self.results['tilt']
        if self.azimuth_true_value is not None and self.azimuth_input is None:
            self.results['azimuth residual'] = self.azimuth_true_value - self.results['azimuth']
        return

    def get_day_range(self, input_data, interval):
        """
        This method was intended to evaluate different day ranges for the estimation of tilt and  azimuth. However, no
        gain was seen from using day ranges instead of the full year. Therefore, the study is usually run with
        `interval=None'
        :param input_data: boolean array containing daytime hours
        :param interval:  day interval to be used in estimation
        :return: Boolean DataFrame with day selection
        """
        if interval is not None:
            day_range = (self.day_of_year > interval[0]) & (self.day_of_year < interval[1])
        else:
            day_range = np.ones(self.day_of_year.shape, dtype=bool)
        output = input_data * self.clear_index * day_range
        return output

    def create_results_table(self):
        cols = ['day range', 'declination method', 'cvx parameter', 'threshold quantile', 'latitude initial value',
                'tilt initial value', 'azimuth initial value']
        if self.lat_input is None:
            cols.append('latitude')
        if self.tilt_input is None:
            cols.append('tilt')
        if self.azimuth_input is None:
            cols.append('azimuth')
        self.results = pd.DataFrame(columns=cols)



from scipy.optimize import curve_fit
import numpy as np

""" Angle of incidence Module
This module contains the function for the calculation of the system's latitude, tilt and azimuth when some of them are
given as precalculates and others are left as unknowns. Unknowns are calculated via fit to the angle of incidence 
cos(theta) equation (1.6.2) in:
Duffie, John A., and William A. Beckman. Solar engineering of thermal processes. New York: Wiley, 1991.
"""


def run_curve_fit(func, keys,  delta, omega, costheta, boolean_filter, init_values, fit_bounds):
    """
    :param func: Angle of incidence model function.
    :param keys: Dynamic keys of parameters being calculated as returned by determine_unknowns.
    :param delta: System's declination in Degrees (array).
    :param omega: System's hour angle in Degrees(array).
    :param costheta: The dependent data. Angle of incidence array used to fit parameters.
    :param boolean_filter: boolean array specifying days to be used in fitting.
    :param init_values: Initial guess for the parameters. (Degrees).
    :param fit_bounds: Lower and upper bounds on parameters.
    :return: Optimal values for the parameters.
    """
    costheta_fit = costheta[boolean_filter]
    x = np.array([np.deg2rad(delta), np.deg2rad(omega)])

    popt, pcov = curve_fit(func, x, costheta_fit, p0=np.deg2rad(init_values), bounds=fit_bounds)

    if 'azimuth_estimate' in keys:
        popt[-1] -= np.rint(popt[-1] / 2 / np.pi) * 2 * np.pi

    estimates = np.degrees(popt)
    return estimates



import numpy as np
import cvxpy as cvx


def filter_data(data_matrix, daytime_threshold=None, x1=None, x2=None):
    """
    :param data_matrix: Pandas DataFrame data matrix with input signal.
    :param daytime_threshold: Optional. Daytime threshold signal value separating day from night.
    :param x1: Float. Parameter for signal decomposition threshold quantile seasonality calculation.
    :param x2: Float. Parameter for signal decomposition threshold quantile seasonality calculation.
    :return: Boolean DataFrame with daylight hours.
    """
    if daytime_threshold is None:
        daytime_threshold_fit = find_daytime_threshold_quantile_seasonality(data_matrix, x1, x2)
        boolean_daytime = data_matrix > daytime_threshold_fit
    else:
        boolean_daytime = data_matrix > daytime_threshold
    return boolean_daytime


def find_daytime_threshold_quantile_seasonality(data_matrix, p1, p2):
    m = cvx.Parameter(nonneg=True, value=10 ** 6)
    # setting local quantile for 10% of the data
    t = cvx.Parameter(nonneg=True, value=p1)
    y = np.quantile(data_matrix, p2, axis=0)
    x1 = cvx.Variable(len(y))
    x2 = cvx.Variable(len(y))
    if data_matrix.shape[1] > 365:
        constraints = [
            x2[365:] == x2[:-365], x1 + x2 == y
        ]
    else:
        constraints = [x1 + x2 == y]
    c1 = cvx.sum(1 / 2 * cvx.abs(x1) + (t - 1 / 2) * x1)
    c2 = cvx.sum_squares(cvx.diff(x2, 2))
    objective = cvx.Minimize(c1 + m * c2)
    prob = cvx.Problem(objective, constraints=constraints)
    prob.solve(solver='MOSEK')
    return x2.value



""" Equation of Time Module
The Equation of Time (EoT) describes the discrepancy between clock time and
solar time: https://en.wikipedia.org/wiki/Equation_of_time

This module contains two approaches to calculating the EoT: da_rosa and Duffie
"""
import numpy as np


def eot_da_rosa(day_of_year):
    """
    The equation of time as defined in:
        Haghdadi, Navid, et al. "A method to estimate the location and
        orientation of distributed photovoltaic systems from their generation
        output data." Renewable Energy 108 (2017): 390-400.

    These are equations (7) and (8) in the paper.

    :param day_of_year: the day of year, can be int, float, or numpy array
    :return: the difference between clock time and solar time for a given day of year
    """
    b = np.deg2rad((360 / 365) * (day_of_year - 81))
    eot = 9.87 * np.sin(2 * b) - 7.53 * np.cos(b) - 1.5 * np.sin(b)
    try:
        return eot.values
    except AttributeError:
        return eot


def eot_duffie(day_of_year):
    """
    The equation of time as defined in:
        Duffie, John A., and William A. Beckman. Solar engineering of thermal
        processes. New York: Wiley, 1991.

    These are equations (1.4.2) and (1.5.3) in the book

    :param day_of_year: the day of year, can be int, float, or numpy array
    :return: the difference between clock time and solar time for a given day of year
    """
    b = np.deg2rad((360 / 365) * (day_of_year - 1))
    A = 1440 / (2 * np.pi)  # book uses approximation of 229.2
    eot = A * (0.000075 + 0.001868 * np.cos(b) - 0.032077 * np.sin(b)
               - 0.014615 * np.cos(2 * b) - 0.04089 * np.sin(2 * b))
    try:
        return eot.values
    except AttributeError:
        return eot



import urllib3 as urllib2
from pykml import parser
import numpy as np
from haversine import haversine
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import os
import math

def ground_truth_estimate(filename):
    def readCoordinates_onepolygon(child):
        coordinates = child.Polygon.outerBoundaryIs.LinearRing.coordinates
        s = coordinates.__str__()
        s = s.split()
        # Extract coordinates
        j = 0
        for i in s:
            if j==1:
                c1 = np.fromstring(i,dtype=np.float,sep=',')
            elif j==2:
                c2 = np.fromstring(i,dtype=np.float,sep=',')
            elif j==3:
                c3 = np.fromstring(i,dtype=np.float,sep=',')
            j = j+1
        return c1,c2,c3

    def coordinatesToCartesian(c1,c2,c3):
        # Convert coordinates to Cartesian
        temp1 = (c2[1],c1[0])
        temp2 = (c3[1],c1[0])
        c1z = c1[2]/1000.0
        c1 = (c1[1],c1[0])
        c2z = c2[2]/1000.0
        c2 = (c2[1],c2[0])
        c3z = c3[2]/1000.0
        c3 = (c3[1],c3[0])
        y1 = haversine(c1, temp1)
        x1 = haversine(c2, temp1)
        y2 = haversine(c1, temp2)
        x2 = haversine(c3, temp2)
        p1 = np.array([0, 0, c1z])
        p2 = np.array([x1, y1, c2z])
        p3 = np.array([x2, y2, c3z])

        return p1,p2,p3

    def normalVec(p1,p2,p3):
        # These two vectors are in the plane
        v1 = p2 - p1
        v2 = p3 - p1

        # The cross product is a vector normal to the plane
        cp = np.cross(v1, v2)
        x, y, z = cp

        #Ensure vector is in the positive direction
        if z<0:
            z = abs(z)
            x = 0-x
            y = 0-y

        return x,y,z

    def tiltAz(x,y,z):
        # Calculate tilt and azimuth
        #tilt = (np.arctan(np.sqrt((x*x)+(y*y))/z))*(180/np.pi)
        tilt = np.degrees(np.arctan2(np.sqrt((x*x)+(y*y))*(180/np.pi), z*(180/np.pi)))
        #azimuth = 90-math.degrees(np.arctan2(y*(180/np.pi), x*(180/np.pi)))
        #azimuth = 90-(np.arctan((y/x))*(180/np.pi))
        #azimuth = 90-math.degrees(np.arctan((y/x)))
        azimuth = 90-np.degrees(np.arctan2(y*(180/np.pi), x*(180/np.pi)))
        if azimuth < -90:
            azimuth = +180 + azimuth
        elif azimuth > 90:
            azimuth = +180 - azimuth
        return tilt,azimuth
    file = filename
    tilt = []
    azimuth = []
    with open(file, 'rb') as f:
        root = parser.fromstring(f.read())
    ch = root.Document.Folder.getchildren()
    for child in ch:
        try:
            c1,c2,c3 = readCoordinates_onepolygon(child)
            #convert coordinates to cartesian
            p1,p2,p3 = coordinatesToCartesian(c1,c2,c3)
            #get the normal vector to the plane
            x,y,z = normalVec(p1,p2,p3)
            #aa.append(x)
            #bb.append(y)
            #cc.append(z)
            #calculate tilt and azimuth
            t,a = tiltAz(x,y,z)
            #gather all tilts and azimuths of same home
            tilt.append(t)
            azimuth.append(a)
        except:
            pass
    #calculate mean and standard deviation tilt and azimuth per home
    ave_tilt = np.mean(tilt)
    med_tilt = np.median(tilt)
    std_tilt = np.std(tilt)
    ave_azimuth = np.mean(azimuth)
    med_azimuth = np.median(azimuth)
    std_azimuth = np.std(azimuth)

    return (ave_tilt, med_tilt, std_tilt, ave_azimuth, med_azimuth, std_azimuth)



"""The function cos(theta) is  calculated using equation (1.6.2) in:
 Duffie, John A., and William A. Beckman. Solar engineering of thermal
 processes. New York: Wiley, 1991."""

import numpy as np


def func_costheta(x, phi, beta, gamma):
    delta = x[0]
    omega = x[1]

    gamma -= np.rint(gamma / 2 / np.pi) * 2 * np.pi

    a = np.sin(delta) * np.sin(phi) * np.cos(beta)
    b = np.sin(delta) * np.cos(phi) * np.sin(beta) * np.cos(gamma)
    c = np.cos(delta) * np.cos(phi) * np.cos(beta) * np.cos(omega)
    d = np.cos(delta) * np.sin(phi) * np.sin(beta) * np.cos(gamma) * np.cos(omega)
    e = np.cos(delta) * np.sin(beta) * np.sin(gamma) * np.sin(omega)
    return a - b + c + d + e



import numpy as np
import cvxpy as cvx


def fit_longitude(eot, solarnoon, days, gmt_offset, loss='l2'):
    lon = cvx.Variable()
    if loss == 'l2':
        cost_func = cvx.norm
    elif loss == 'l1':
        cost_func = cvx.norm1
    elif loss == 'huber':
        cost_func = lambda x: cvx.sum(cvx.huber(x))

    sn_m = 720 - eot + 4 * (15 * gmt_offset - lon)
    sn_h = sn_m / 60
    nan_mask = np.isnan(solarnoon)
    use_days = np.logical_and(days, ~nan_mask)
    cost = cost_func(sn_h[use_days] - solarnoon[use_days])
    objective = cvx.Minimize(cost)
    problem = cvx.Problem(objective)
    problem.solve()
    return lon.value.item()



from solardatatools.algorithms import SunriseSunset


def get_optimized_sunrise_sunset(filled_data_matrix=None, raw_data_matrix=None):
    optimized_dict = {}
    ss = SunriseSunset()
    if raw_data_matrix is not None:
        ss.run_optimizer(data=raw_data_matrix)
        optimized_dict['est_sr_raw'] = ss.sunrise_estimates
        optimized_dict['est_ss_raw'] = ss.sunset_estimates
        optimized_dict['meas_sr_raw'] = ss.sunrise_measurements
        optimized_dict['meas_ss_raw'] = ss.sunset_measurements
        optimized_dict['thres_raw'] = ss.threshold
    else:
        optimized_dict['est_sr_raw'] = None
        optimized_dict['est_ss_raw'] = None
        optimized_dict['meas_sr_raw'] = None
        optimized_dict['meas_ss_raw'] = None
        optimized_dict['thres_raw'] = None

    if filled_data_matrix is not None:
        ss.run_optimizer(data=filled_data_matrix)
        optimized_dict['est_sr_f'] = ss.sunrise_estimates
        optimized_dict['est_ss_f'] = ss.sunset_estimates
        optimized_dict['meas_sr_f'] = ss.sunrise_measurements
        optimized_dict['meas_ss_f'] = ss.sunset_measurements
        optimized_dict['thres_f'] = ss.threshold
    else:
        optimized_dict['est_sr_f'] = None
        optimized_dict['est_ss_f'] = None
        optimized_dict['meas_sr_f'] = None
        optimized_dict['meas_ss_f'] = None
        optimized_dict['thres_f'] = None
    return optimized_dict



import numpy as np
import cvxpy as cvx

"""
Calculates the angle incidence for a system based on its power matrix using signal decomposition.
"""


def find_fit_costheta(data_matrix, clear_index):
    """
    :param data_matrix: power matrix.
    :param clear_index: boolean array specifying clear days.
    :return: angle of incidence array.
    """
    data = np.max(data_matrix, axis=0)
    s1 = cvx.Variable(len(data))
    s2 = cvx.Variable(len(data))
    cost = 1e1 * cvx.norm(cvx.diff(s1, k=2), p=2) + cvx.norm(s2[clear_index])
    objective = cvx.Minimize(cost)
    constraints = [
        data == s1 + s2,
        s1[365:] == s1[:-365]
    ]
    problem = cvx.Problem(objective, constraints)
    problem.solve(solver='MOSEK')
    scale_factor_costheta = s1.value
    costheta_fit = data_matrix / np.max(s1.value)
    return scale_factor_costheta, costheta_fit



"""
Longitude Study Module
This module contains a class for conducting a study of different approaches to estimating longitude from solar data.
This code accepts solar power data in the form of a `solar-data-tools` `DataHandler` object, which is used to
standardize and pre-process the data. The provided class will then estimate the longitude of the site that produced the
 data, using configurations that can be set in the `run` method. The basic concept is to estimate solar noon for each
 day based on the measured data, and then use the relationship between standard time, solar time, and the equation of
 time to estimate the longitude.
The following configurations can be run:

 - Input data matrix: 'raw', 'filled'
 - Equation of time (EoT) estimator: Duffie or Da Rosa
 - Estimation algorithm: calculation from EoT definition, curve fitting with
   L2 loss, curve fitting with L1 loss, or curve fitting with Huber loss
 - Method for solar noon estimation: average of sunrise and sunset, the energy center of mass, optimized estimates,
   optimized measurements.
 - Method for day selection: all days, sunny/clear days, cloudy days

"""
import numpy as np
import pandas as pd
from solardatatools.solar_noon import energy_com, avg_sunrise_sunset
from pvsystemprofiler.utilities.equation_of_time import eot_da_rosa, eot_duffie
from pvsystemprofiler.utilities.progress import progress
from pvsystemprofiler.algorithms.longitude.estimation import estimate_longitude
from pvsystemprofiler.algorithms.optimized_sunrise_sunset import get_optimized_sunrise_sunset


class LongitudeStudy():
    def __init__(self, data_handler, gmt_offset=-8, true_value=None):
        """
        Default value for GMT offset is -8 which corresponds to Pacific
        Standard Time, or systems located in California.
        :param data_handler: `DataHandler` class instance loaded with a solar power data set
        :param gmt_offset: The offset in hours between the local timezone and GMT/UTC
        :param true_value: (optional) the ground truth value for the system's longitude
        """
        self.data_handler = data_handler
        if not data_handler._ran_pipeline:
            print('Running DataHandler preprocessing pipeline with defaults')
            self.data_handler.run_pipeline()
        self.data_matrix = self.data_handler.filled_data_matrix
        self.raw_data_matrix = self.data_handler.raw_data_matrix
        self.true_value = true_value
        self.opt_threshold_raw = None
        self.opt_threshold_filled = None
        # Attributes used for all calculations
        self.gmt_offset = gmt_offset
        self.day_of_year = self.data_handler.day_index.dayofyear
        self.eot_duffie = eot_duffie(self.day_of_year)
        self.eot_da_rosa = eot_da_rosa(self.day_of_year)
        # Attributes that change depending on the configuration
        self.solarnoon = None
        self.days = None
        # Results
        self.results = None
        self.best_result = None
        self.estimates_sunrise_raw = None
        self.estimates_sunset_raw = None
        self.measurements_sunrise_raw = None
        self.measurements_sunset_raw = None
        self.estimates_sunrise_filled = None
        self.estimates_sunset_filled = None
        self.measurements_sunrise_filled = None
        self.measurements_sunset_filled = None

    def run(self, data_matrix=('raw', 'filled'),
            estimator=('calculated', 'fit_l1', 'fit_l2', 'fit_huber'),
            eot_calculation=('duffie', 'da_rosa'),
            solar_noon_method=('rise_set_average', 'energy_com', 'optimized_estimates', 'optimized_measurements'),
            day_selection_method=('all', 'clear', 'cloudy'),
            verbose=True):
        """
        Run a study with the given configuration of options. Defaults to
        running all available options. Any kwarg can be constrained by
        providing a subset of acceptable keys. For example the default keys
        for the estimator kwarg are:

            ('calculated', 'fit_l1', 'fit_l2', 'fit_huber')

        Additionally, any of the following would be acceptable for this kwarg:

            ('calculated', 'fit_l1', 'fit_l2', 'fit_huber')
            ('fit_l2', 'fit_huber')
            ('fit_l2',)
            'fit_l2'

        This method sets the `results` attribute to be a pandas data frame
        containing the results of the study. If a ground truth value was
        provided to the class constructor, the best result will be assigned
        to the `best_result` attribute.
        :param data_matrix: 'raw', 'filled'.
        :param estimator: 'calculated', 'fit_l1', 'fit_l2', 'fit_huber'.
        :param eot_calculation: 'duffie', 'da_rosa'.
        :param solar_noon_method: 'rise_set_average', 'energy_com', 'optimized_estimates', 'optimized_measurements'.
        :param day_selection_method: 'all', 'clear', 'cloudy'.
        :param verbose: show progress bar if True.
        :return: None.
        """
        results = pd.DataFrame(columns=[
            'longitude', 'estimator', 'eot_calculation', 'solar_noon_method',
            'day_selection_method', 'data_matrix'
        ])
        estimator = np.atleast_1d(estimator)
        eot_calculation = np.atleast_1d(eot_calculation)
        solar_noon_method = np.atleast_1d(solar_noon_method)
        day_selection_method = np.atleast_1d(day_selection_method)
        data_matrix = np.atleast_1d(data_matrix)

        if 'raw' in data_matrix:
            rdm = self.raw_data_matrix
        else:
            rdm = None
        if 'filled' in data_matrix:
            fdm = self.data_matrix
        else:
            fdm = None
        opt_dict = get_optimized_sunrise_sunset(fdm, rdm)
        self.estimates_sunrise_raw, self.estimates_sunset_raw, self.measurements_sunrise_raw, \
        self.measurements_sunset_raw, self.opt_threshold_raw, \
        self.estimates_sunrise_filled, self.estimates_sunset_filled, self.measurements_sunrise_filled, \
        self.measurements_sunset_filled, self.opt_threshold_filled = opt_dict.values()

        total = (len(estimator) * len(eot_calculation) * len(solar_noon_method)
                 * len(day_selection_method) * len(data_matrix))
        counter = 0
        for dm in data_matrix:
            if dm == 'raw':
                data_in = self.raw_data_matrix
            elif dm == 'filled':
                data_in = self.data_matrix
            for sn in solar_noon_method:
                if sn == 'rise_set_average':
                    self.solarnoon = avg_sunrise_sunset(data_in)
                elif sn == 'energy_com':
                    self.solarnoon = energy_com(data_in)
                elif sn == 'optimized_estimates':
                    if dm == 'filled':
                        sunset = np.copy(self.estimates_sunset_filled)
                        sunrise = np.copy(self.estimates_sunrise_filled)
                    if dm == 'raw':
                        sunset = np.copy(self.estimates_sunset_raw)
                        sunrise = np.copy(self.estimates_sunrise_raw)
                    self.solarnoon = np.nanmean([sunrise, sunset], axis=0)
                elif sn == 'optimized_measurements':
                    if dm == 'filled':
                        sunset = np.copy(self.measurements_sunset_filled)
                        sunrise = np.copy(self.measurements_sunrise_filled)
                    if dm == 'raw':
                        sunset = np.copy(self.measurements_sunset_raw)
                        sunrise = np.copy(self.measurements_sunrise_raw)
                    sunrise[np.isnan(sunrise)] = 0
                    sunset[np.isnan(sunset)] = 0
                    self.solarnoon = np.nanmean([sunrise, sunset], axis=0)
                for ds in day_selection_method:
                    if ds == 'all':
                        self.days = self.data_handler.daily_flags.no_errors
                    elif ds == 'clear':
                        self.days = self.data_handler.daily_flags.clear
                    elif ds == 'cloudy':
                        self.days = self.data_handler.daily_flags.cloudy
                    for est in estimator:
                        for eot in eot_calculation:
                            if verbose:
                                progress(counter, total)

                            try:
                                if eot in ('duffie', 'd', 'duf') or eot is None:
                                    eot_ref = self.eot_duffie
                                elif eot in ('da_rosa', 'dr', 'rosa'):
                                    eot_ref = self.eot_da_rosa
                                lon = estimate_longitude(est, eot_ref, self.solarnoon, self.days, self.gmt_offset)

                            except ValueError:
                                lon = np.nan

                            results.loc[counter] = [
                                lon, est, eot, sn, ds, dm
                            ]
                            counter += 1
        if verbose:
            progress(counter, total)
        if self.true_value is not None:
            results['residual'] = self.true_value - results['longitude']
            results['measured_longitude'] = self.true_value
        self.results = results
        if self.true_value is not None:
            best_loc = results['residual'].apply(lambda x: np.abs(x)).argmin()
            self.best_result = results.loc[best_loc]
            self.results = results.loc[np.argsort(np.abs(results['residual']).values)]
        return



"""
Creates and runs local partitions scripts for run_partition_scripts
"""
import sys
import os
import pandas as pd

# Input parameters
start_index = int(sys.argv[1])
end_index = int(sys.argv[2])
script_name = str(sys.argv[3])
estimation = str(sys.argv[4])
global_input_file = str(sys.argv[5])
local_working_folder = str(sys.argv[6])
local_input_file = str(sys.argv[7])
local_output_file = str(sys.argv[8])
power_column_id = str(sys.argv[9])
convert_to_ts = str(sys.argv[10])
s3_location = str(sys.argv[11])
n_files = str(sys.argv[12])
file_label = str(sys.argv[13])
fix_time_shifts = str(sys.argv[14])
time_zone_correction = str(sys.argv[15])
check_json = str(sys.argv[16])
gmt_offset = str(sys.argv[17])
data_type = str(sys.argv[18])
supplementary_file = str(sys.argv[19])
python_command = str(sys.argv[20])

# read full list of systems save a local copy of systems corresponding to partition
df_full = pd.read_csv(global_input_file, index_col=0)
df_part = df_full.copy()
df_part = df_part[start_index:end_index]
df_part.to_csv(local_input_file)

# create execute command to run study or site report
command = 'setsid nohup' + ' ' + python_command + ' ' + script_name
# create arguments to run `command` with
arguments = estimation + ' ' \
            + local_input_file + ' '  \
            + n_files + ' ' \
            + s3_location + ' ' \
            + file_label + ' ' \
            + power_column_id + ' ' \
            + local_output_file + ' ' \
            + fix_time_shifts + ' '\
            + time_zone_correction + ' '\
            + check_json + ' ' \
            + convert_to_ts + ' ' \
            + supplementary_file + ' ' \
            + gmt_offset + ' ' \
            + data_type

full_command = command + ' ' + arguments + '>out &'
# save local copy of run script
file1 = open(local_working_folder + 'run_local_partition.sh', "w")
print('Running local script')
file1.write('#!/bin/sh\n')
file1.write(full_command)
file1.close()

# execute local run script
os.system(full_command)



""" Longitude run script
This run script allows to run the longitude_study for multiple sites. The site ids to be evaluated can be provided in
 a csv file. Alternatively, the path to a folder containing the input signals of the sites in separate csv files can be
 provided.  The script provides the option to provided the full path to csv file containing latitude and gmt offset for
 each system for comparison.
"""
import sys
from pathlib import Path
import pandas as pd
import numpy as np
from time import time

# TODO: remove pth.append after package is deployed
filepath = Path(__file__).resolve().parents[2]
sys.path.append(str(filepath))
from solardatatools.utilities import progress
from pvsystemprofiler.scripts.modules.script_functions import run_failsafe_pipeline
from pvsystemprofiler.scripts.modules.script_functions import resume_run
from pvsystemprofiler.scripts.modules.script_functions import load_generic_data
from pvsystemprofiler.scripts.modules.script_functions import log_file_versions
from pvsystemprofiler.scripts.modules.script_functions import load_system_metadata
from pvsystemprofiler.scripts.modules.script_functions import generate_list
from solardatatools.dataio import load_cassandra_data
from pvsystemprofiler.scripts.modules.script_functions import extract_sys_parameters
from pvsystemprofiler.scripts.modules.script_functions import get_commandline_inputs
from pvsystemprofiler.scripts.modules.script_functions import run_failsafe_lon_estimation
from pvsystemprofiler.scripts.modules.script_functions import run_failsafe_lat_estimation
from pvsystemprofiler.scripts.modules.script_functions import run_failsafe_ta_estimation
from solardatatools import DataHandler


def evaluate_systems(site_id, inputs_dict, df, site_metadata, json_file_dict=None):
    partial_df_cols = ['site', 'system', 'passes pipeline', 'length', 'capacity_estimate', 'data_sampling',
                       'data quality_score', 'data clearness_score', 'inverter_clipping', 'time_shifts_corrected',
                       'time_zone_correction', 'capacity_changes', 'normal_quality_scores']

    if json_file_dict is not None:
        partial_df_cols.extend(['zip_code', 'real longitude', 'real latitude', 'real tilt', 'real azimuth'])
    if inputs_dict['time_shift_manual']:
        partial_df_cols.append('time_shift_manual')

    partial_df = pd.DataFrame(columns=partial_df_cols)

    ll = len(inputs_dict['power_column_label'])

    if inputs_dict['convert_to_ts']:
        dh = DataHandler(df, convert_to_ts=inputs_dict['convert_to_ts'])
        cols = [el[-1] for el in dh.keys]
    else:
        cols = df.columns

    i = 0
    for col_label in cols:
        if col_label.find(inputs_dict['power_column_label']) != -1:
            system_id = col_label[ll:]
            if system_id in site_metadata['system'].tolist():
                i += 1
                dh = DataHandler(df, convert_to_ts=inputs_dict['convert_to_ts'])
                sys_tag = inputs_dict['power_column_label'] + system_id
                sys_mask = site_metadata['system'] == system_id

                if inputs_dict['time_shift_manual']:
                    time_shift_manual = int(site_metadata.loc[sys_mask, 'time_shift_manual'].values[0])
                    if time_shift_manual == 1:
                        dh.fix_dst()
                else:
                    time_shift_manual = 0

                dh, passes_pipeline = run_failsafe_pipeline(dh, sys_tag, inputs_dict['fix_time_shifts'],
                                                            inputs_dict['time_zone_correction'])
                if passes_pipeline:
                    results_list = [site_id, system_id, passes_pipeline, dh.num_days, dh.capacity_estimate,
                                    dh.data_sampling, dh.data_quality_score, dh.data_clearness_score,
                                    dh.inverter_clipping, dh.time_shifts, dh.tz_correction, dh.capacity_changes,
                                    dh.normal_quality_scores]

                    if inputs_dict['time_shift_manual']:
                        results_list.append(time_shift_manual)
                    if json_file_dict is not None:
                        if system_id in json_file_dict.keys():
                            source_file = json_file_dict[system_id]
                            json_information = extract_sys_parameters(source_file, system_id,
                                                                      inputs_dict['s3_location'])
                        else:
                            json_information = [np.nan] * 4
                        results_list.extend(json_information)

                else:
                    results_list = [site_id, system_id, passes_pipeline] + [np.nan] * (len(results_list) - 3)

                if inputs_dict['estimation'] == 'longitude':
                    if inputs_dict['longitude']:
                        real_longitude = float(site_metadata.loc[sys_mask, 'longitude'])
                    if inputs_dict['gmt_offset'] is not None:
                        gmt_offset = inputs_dict['gmt_offset']
                    else:
                        gmt_offset = float(site_metadata.loc[sys_mask, 'gmt_offset'])
                    results_df, passes_estimation = run_failsafe_lon_estimation(dh, real_longitude, gmt_offset)

                elif inputs_dict['estimation'] == 'latitude':
                    if inputs_dict['latitude']:
                        real_latitude = float(site_metadata.loc[sys_mask, 'latitude'])
                    results_df, passes_estimation = run_failsafe_lat_estimation(dh, real_latitude)

                elif inputs_dict['estimation'] == 'tilt_azimuth':
                    if inputs_dict['estimated_longitude']:
                        longitude_input = float(site_metadata.loc[sys_mask, 'estimated_longitude'])
                    if inputs_dict['estimated_latitude']:
                        latitude_input = float(site_metadata.loc[sys_mask, 'latitude'])
                    if inputs_dict['latitude']:
                        real_latitude = float(site_metadata.loc[sys_mask, 'latitude'])
                    if inputs_dict['tilt']:
                        real_tilt = float(site_metadata.loc[sys_mask, 'tilt'])
                    if inputs_dict['azimuth']:
                        real_azimuth = float(site_metadata.loc[sys_mask, 'azimuth'])
                    if inputs_dict['gmt_offset']:
                        gmt_offset = inputs_dict['gmt_offset']
                    else:
                        gmt_offset = float(site_metadata.loc[sys_mask, 'gmt_offset'])
                    results_df, passes_estimation = run_failsafe_ta_estimation(dh, 1, None, longitude_input,
                                                                               latitude_input, None, None,
                                                                               real_latitude, real_tilt, real_azimuth,
                                                                               gmt_offset)
                if inputs_dict['estimation'] in ['longitude', 'latitude', 'tilt_azimuth']:
                    results_df[partial_df_cols] = results_list
                    partial_df = partial_df.append(results_df)
                elif inputs_dict['estimation'] == 'report':
                    partial_df.loc[len(partial_df)] = results_list
    return partial_df


def main(full_df, inputs_dict, df_system_metadata):
    site_run_time = 0
    total_time = 0
    file_list, json_file_dict = generate_list(inputs_dict, full_df, df_system_metadata)

    if inputs_dict['n_files'] != 'all':
        file_list = file_list[:int(inputs_dict['n_files'])]
    if full_df is None:
        full_df = pd.DataFrame()

    for file_ix, file_id in enumerate(file_list):
        t0 = time()
        msg = 'Site/Accum. run time: {0:2.2f} s/{1:2.2f} m'.format(site_run_time, total_time / 60.0)
        progress(file_ix, len(file_list), msg, bar_length=20)

        if inputs_dict['file_label'] is not None:
            i = file_id.find(inputs_dict['file_label'])
            site_id = file_id[:i]
            mask = df_system_metadata['site'] == site_id.split(inputs_dict['file_label'])[0]
        else:
            site_id = file_id.split('.')[0]

            mask = df_system_metadata['site'] == site_id
        site_metadata = df_system_metadata[mask]

        # TODO: integrate option for other data inputs
        if inputs_dict['data_source'] == 's3':
            df = load_generic_data(inputs_dict['s3_location'], inputs_dict['file_label'], site_id)
        if inputs_dict['data_source'] == 'cassandra':
            df = load_cassandra_data(site_id)

        if not site_metadata.empty:
            partial_df = evaluate_systems(site_id, inputs_dict, df, site_metadata, json_file_dict)
        else:
            partial_df = None

        if not partial_df.empty or partial_df is not None:
            full_df = full_df.append(partial_df)
            full_df.index = np.arange(len(full_df))
            full_df.to_csv(inputs_dict['output_file'])
            t1 = time()
            site_run_time = t1 - t0
            total_time += site_run_time

    msg = 'Site/Accum. run time: {0:2.2f} s/{1:2.2f} m'.format(site_run_time, total_time / 60.0)
    if len(file_list) != 0:
        progress(len(file_list), len(file_list), msg, bar_length=20)
    print('finished')
    return


if __name__ == '__main__':
    """
    :param estimation: Estimation to be performed. Options are 'report', 'longitude', 'latitude', 'tilt_azimuth'
    :param input_site_file:  csv file containing list of sites to be evaluated. 'None' if no input file is provided.
    :param n_files: number of files to read. If 'all' all files in folder are read.
    :param s3_location: Absolute path to s3 location of files.
    :param file_label:  Repeating portion of data files label. If 'None', no file label is used. 
    :param power_column_label: Repeating portion of the power column label. 
    :param output_file: Absolute path to csv file containing report results.
    :param fix_time_shits: String, 'True' or 'False'. Specifies if time shifts are to be 
    fixed when running the pipeline.
    :param time_zone_correction: String, 'True' or 'False'. Specifies if the time zone correction is performed when 
    running the pipeline.
    :param check_json: String, 'True' or 'False'. Check json file for location information.
    :param convert_to_ts: String, 'True' or 'False'. Specifies if conversion to time series is performed when 
    running the pipeline.
    :param system_summary_file: Full path to csv file containing longitude and manual time shift flag for each system,
    None if no file
    provided. 
    :param gmt_offset: String. Single value of gmt offset to be used for all estimations. If None a list with individual
    gmt offsets needs to be provided.
    :param data_source: String. Input signal data source. Options are 's3' and 'cassandra'.
    """

    input_kwargs = sys.argv
    inputs_dict = get_commandline_inputs(input_kwargs)

    log_file_versions('solar-data-tools', active_conda_env='pvi-user')
    log_file_versions('pv-system-profiler')

    full_df = resume_run(inputs_dict['output_file'])

    ssf = inputs_dict['system_summary_file']
    if ssf is not None:
        df_system_metadata = load_system_metadata(df_in=ssf, file_label=inputs_dict['file_label'])
        cols = df_system_metadata.columns
        for param in ['longitude', 'latitude', 'tilt', 'azimuth',
                      'estimated_longitude', 'estimated_latitude',
                      'time_shift_manual']:
            if param in cols:
                inputs_dict[param] = True
            else:
                inputs_dict[param] = False
    else:
        df_system_metadata = None

main(full_df, inputs_dict, df_system_metadata)



import numpy as np
from solardatatools.sunrise_sunset import rise_set_rough
from solardatatools.daytime import find_daytime, detect_sun


def calculate_hours_daylight_raw(data_in, data_sampling, threshold=0.001):
    boolean_daytime = find_daytime(data_in, threshold)
    return (np.sum(boolean_daytime, axis=0)) * data_sampling / 60


def calculate_hours_daylight(data_in, threshold=0.001):
    bool_msk = detect_sun(data_in, threshold=threshold)
    measurements = rise_set_rough(bool_msk)
    return measurements['sunsets'] - measurements['sunrises']



import numpy as np


def determine_keys(latitude, tilt, azimuth):
    key = []
    if latitude is None:
        key.append('latitude_estimate')
    if tilt is None:
        key.append('tilt_estimate')
    if azimuth is None:
        key.append('azimuth_estimate')
    return key


def select_init_values(bounds_dict, dict_keys):
    init_vals = []
    ivr = []
    if 'latitude_estimate' in dict_keys:
        init_vals.append(bounds_dict['latitude'])
        ivr.append(bounds_dict['latitude'])
    else:
        ivr.append(np.nan)
    if 'tilt_estimate' in dict_keys:
        init_vals.append(bounds_dict['tilt'])
        ivr.append(bounds_dict['tilt'])
    else:
        ivr.append(np.nan)
    if 'azimuth_estimate' in dict_keys:
        init_vals.append(bounds_dict['azimuth'])
        ivr.append(bounds_dict['azimuth'])
    else:
        ivr.append(np.nan)
    return init_vals, ivr
